#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import chart_object
from . import color
from . import line_style_doc
from . import object_set
from . import theme
from .pychart_types import *

_keys = {
    "width": (UnitType, theme.default_line_width, "Width of the line, in points."),
    "color": (color.T, color.default, "The color of the line."),
    "dash": (TupleType, None,
             """The value
             of None will draw a solid line. Otherwise, this
             attribute specifies the style of dashed lines. 
             The 2N'th value specifies the length of the line (in points), 
             and 2N+1'th value specifies the length of the blank.

             For example, the dash style of (3,2,4,1) draws a dashed line that
             looks like @samp{---__----_---__----_...}.
             """),
    "cap_style": (IntType, 0,
                  """Defines the style of the tip of the line segment.
                  0: butt cap (square cutoff, with no projection beyond),
                  1: round cap (arc), 2: projecting square cap
                  (square cutoff, but the line extends half the line width).
                  See also Postscript/PDF reference manual."""),
    "join_style": (IntType, 0,
                   """Join style. 0: Miter join (sharp, pointed corners),
                   1: round join (rounded corners),
                   2: bevel join (flattened corners).
                   See also Postscript/PDF reference manual.""")
}


class T(chart_object.T):
    __doc__ = line_style_doc.doc
    keys = _keys

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def __str__(self):
        s = name_table().lookup(self)
        if s:
            return s
        return "<linestyle: width=%s, color=%s, dash=%s, cap=%d, join=%d>" \
               % (self.width, self.color, self.dash, self.cap_style, self.join_style)


default = T(color=color.default)

dash1 = 1.5, 1.5  # - - - -
dash2 = 5, 2, 5, 2  # -- -- -- --
dash3 = 1, 1

black = T(color=color.black)
black_dash1 = T(color=color.black, dash=dash1)
black_dash2 = T(color=color.black, dash=dash2)
black_dash3 = T(color=color.black, dash=dash3)

gray70 = T(color=color.gray70)
gray70_dash1 = T(color=color.gray70, dash=dash1)
gray70_dash2 = T(color=color.gray70, dash=dash2)
gray70_dash3 = T(color=color.gray70, dash=dash3)

gray10 = T(color=color.gray10)
gray10_dash1 = T(color=color.gray10, dash=dash1)
gray10_dash2 = T(color=color.gray10, dash=dash2)
gray10_dash3 = T(color=color.gray10, dash=dash3)

gray50 = T(color=color.gray50)
gray50_dash1 = T(color=color.gray50, dash=dash1)
gray50_dash2 = T(color=color.gray50, dash=dash2)
gray50_dash3 = T(color=color.gray50, dash=dash3)

gray60 = T(color=color.gray60)
gray60_dash1 = T(color=color.gray60, dash=dash1)
gray60_dash2 = T(color=color.gray60, dash=dash2)
gray60_dash3 = T(color=color.gray60, dash=dash3)

gray90 = T(color=color.gray90)
gray90_dash1 = T(color=color.gray90, dash=dash1)
gray90_dash2 = T(color=color.gray90, dash=dash2)
gray90_dash3 = T(color=color.gray90, dash=dash3)

gray30 = T(color=color.gray30)
gray30_dash1 = T(color=color.gray30, dash=dash1)
gray30_dash2 = T(color=color.gray30, dash=dash2)
gray30_dash3 = T(color=color.gray30, dash=dash3)

white = T(color=color.white)
default = black

red = T(color=color.red)
darkblue = T(color=color.darkblue)
darkseagreen = T(color=color.darkseagreen)
darkkhaki = T(color=color.darkkhaki)

blue = T(color=color.blue)
green = T(color=color.green)

red_dash1 = T(color=color.red, dash=dash1)
darkblue_dash1 = T(color=color.darkblue, dash=dash1)
darkseagreen_dash1 = T(color=color.darkseagreen, dash=dash1)
darkkhaki_dash1 = T(color=color.darkkhaki, dash=dash1)

red_dash2 = T(color=color.red, dash=dash2)
darkblue_dash2 = T(color=color.darkblue, dash=dash2)
darkseagreen_dash2 = T(color=color.darkseagreen, dash=dash2)
darkkhaki_dash2 = T(color=color.darkkhaki, dash=dash2)

standards = None
_name_table = None


def init():
    global standards, _name_table
    standards = object_set.T()

    if theme.use_color:
        standards.add(black, red, darkblue, gray70, darkseagreen,
                      darkkhaki, gray30,
                      black_dash1, red_dash1, darkblue_dash1, gray70_dash1,
                      darkseagreen_dash1, darkkhaki_dash1, gray30_dash1,
                      black_dash2, red_dash2, darkblue_dash2, gray70_dash2,
                      darkseagreen_dash2, darkkhaki_dash2, gray30_dash2)
    else:
        standards.add(black, black_dash1, black_dash2,
                      gray70, gray70_dash1, gray70_dash2,
                      gray10, gray10_dash1, gray10_dash2,
                      gray50, gray50_dash1, gray50_dash2,
                      gray90, gray90_dash1, gray90_dash2,
                      gray30, gray30_dash1, gray30_dash2,
                      black_dash3,
                      gray70_dash3, gray10_dash3, gray50_dash3, gray90_dash3)
    for style in standards.list():
        style.width = theme.default_line_width
    _name_table = None


def name_table():
    global _name_table
    if not _name_table:
        _name_table = pychart_util.symbol_lookup_table(globals(), standards)
    return _name_table


init()
theme.add_reinitialization_hook(init)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import chart_object
from . import error_bar
from . import legend
from . import line_plot_doc
from . import line_style
from . import object_set
from . import theme
from . import tick_mark
from . import pychart_util
from .pychart_types import *

default_width = 1.2
line_style_itr = iter([])  # type: ignore

_keys = {
    'data': (AnyType, None, pychart_util.data_desc),
    'label': (StringType, '???', pychart_util.label_desc),
    'data_label_offset': (CoordType, (0, 5),
                          """The location of data labels relative to the sample point. Meaningful only when data_label_format != None."""),
    'data_label_format': (FormatType, None,
                          """The format string for the label printed 
                          beside a sample point.
                          It can be a `printf' style format string, or 
                          a two-parameter function that takes the (x, y)
                          values and returns a string. """
                          + pychart_util.string_desc),
    'xcol': (IntType, 0, pychart_util.xcol_desc),
    'ycol': (IntType, 1, pychart_util.ycol_desc),
    'y_error_minus_col': (IntType, 2,
                          """The column (within "data") from which the depth of the errorbar is extracted. Meaningful only when error_bar != None. <<error_bar>>"""),
    'y_error_plus_col': (IntType, -1,
                         """The column (within "data") from which the height of the errorbar is extracted. Meaningful only when error_bar != None. <<error_bar>>"""),
    'y_qerror_minus_col': (IntType, -1, '<<error_bar>>'),
    'y_qerror_plus_col': (IntType, -1, '<<error_bar>>'),

    'line_style': (line_style.T, lambda: next(line_style_itr), pychart_util.line_desc,
                   "By default, a style is picked from standard styles round-robin. <<line_style>>"),

    'tick_mark': (tick_mark.T, None, pychart_util.tick_mark_desc),
    'error_bar': (error_bar.T, None,
                  'The style of the error bar. <<error_bar>>'),
}


class T(chart_object.T):
    __doc__ = line_plot_doc.doc
    keys = _keys

    def check_integrity(self):
        assert chart_object.T.check_integrity(self)

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def get_data_range(self, which):
        if which == 'X':
            return pychart_util.get_data_range(self.data, self.xcol)
        else:
            return pychart_util.get_data_range(self.data, self.ycol)

    def get_legend_entry(self):
        if self.label:
            line_style = self.line_style
            if not line_style and self.error_bar:
                line_style = getattr(self.error_bar, 'line_style', None) or \
                             getattr(self.error_bar, 'hline_style', None) or \
                             getattr(self.error_bar, 'vline_style', None)
                if not line_style:
                    raise Exception('Line plot has label, but an empty line style and error bar.')
            return legend.Entry(line_style=line_style,
                                tick_mark=self.tick_mark,
                                fill_style=None,
                                label=self.label)
        return None

    def draw(self, ar, can):

        # Draw the line

        clipbox = theme.adjust_bounding_box([ar.loc[0], ar.loc[1],
                                             ar.loc[0] + ar.size[0],
                                             ar.loc[1] + ar.size[1]])

        can.clip(clipbox[0], clipbox[1], clipbox[2], clipbox[3])
        if self.line_style:
            points = []
            for pair in self.data:
                yval = pychart_util.get_sample_val(pair, self.ycol)
                xval = pair[self.xcol]
                if None not in (xval, yval):
                    points.append((ar.x_pos(xval), ar.y_pos(yval)))
            can.lines(self.line_style, points)
        can.endclip()

        # Draw tick marks and error bars
        can.clip(ar.loc[0] - 10, ar.loc[1] - 10,
                 ar.loc[0] + ar.size[0] + 10,
                 ar.loc[1] + ar.size[1] + 10)
        for pair in self.data:
            x = pair[self.xcol]
            y = pychart_util.get_sample_val(pair, self.ycol)
            if None in (x, y): continue

            x_pos = ar.x_pos(x)
            y_pos = ar.y_pos(y)

            if self.error_bar:
                plus = pair[self.y_error_plus_col or self.y_error_minus_col]
                minus = pair[self.y_error_minus_col or self.y_error_plus_col]
                if self.y_qerror_minus_col or self.y_qerror_plus_col:
                    q_plus = pair[self.y_qerror_plus_col or self.y_qerror_minus_col]
                    q_minus = pair[self.y_qerror_minus_col or self.y_qerror_plus_col]
                    if None not in (minus, plus, q_minus, q_plus):
                        self.error_bar.draw(can, (x_pos, y_pos),
                                            ar.y_pos(y - minus),
                                            ar.y_pos(y + plus),
                                            ar.y_pos(y - q_minus),
                                            ar.y_pos(y + q_plus))
                else:
                    if None not in (minus, plus):  # PDS
                        self.error_bar.draw(can, (x_pos, y_pos),
                                            ar.y_pos(y - minus),
                                            ar.y_pos(y + plus))

            if self.tick_mark:
                self.tick_mark.draw(can, x_pos, y_pos)
            if self.data_label_format:
                can.show(x_pos + self.data_label_offset[0],
                         y_pos + self.data_label_offset[1],
                         '/hC' + pychart_util.apply_format(self.data_label_format, (x, y), 1))

        can.endclip()


def init():
    global line_style_itr
    line_styles = object_set.T()
    for org_style in line_style.standards.list():
        style = line_style.T(width=default_width, color=org_style.color,
                             dash=org_style.dash)
        line_styles.add(style)

    line_style_itr = line_styles.iterate()


theme.add_reinitialization_hook(init)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
#
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import types

from . import pychart_types


def set_defaults(cls, **dict):
    validAttrs = getattr(cls, "keys")
    for attr, val in list(dict.items()):
        if attr not in validAttrs:
            raise Exception("%s: unknown attribute %s." % (cls, attr))
        tuple = list(validAttrs[attr])
        # 0 : type
        # 1: defaultValue
        # 2: document
        # 3: defaultValue document (optional)
        tuple[1] = val
        validAttrs[attr] = tuple


class T(object):
    def init(self, args):
        keys = self.keys
        for attr, tuple in list(keys.items()):
            defaultVal = tuple[1]
            if isinstance(defaultVal, types.FunctionType):
                # if the value is procedure, use the result of the proc call
                # as the default value
                defaultVal = defaultVal(*())
            setattr(self, attr, defaultVal)

        for key, val in list(args.items()):
            self.__setattr__(key, val)

    def __check_type(self, item, value):
        if item.startswith("_"):
            return

        if item not in self.keys:
            raise Exception("%s: unknown attribute '%s'" % (self, item))

        typeval, default_value, docstring = self.keys[item][0:3]
        if value == None or typeval == pychart_types.AnyType:
            pass
        elif typeval == bool:
            # In 2.3, bool is a full-fledged type, whereas
            # in 2.2, it's just a function that returns an integer.
            # To mask this difference, we handle the bool type specially.
            if value not in (True, False):
                raise TypeError("%s: Expecting bool, but got %s" % (self, value))
        elif typeval == str:
            if not isinstance(value, str) and not isinstance(value, str):
                raise TypeError("%s: Expecting a string, but got %s" % (self, value))
        elif isinstance(typeval, types.FunctionType):
            # user-defined check procedure
            error = typeval(*(value,))
            if error != None:
                raise TypeError("%s: %s for attribute '%s', but got '%s'" % (self, error, item, value))
        elif isinstance(value, typeval):
            pass
        else:
            raise TypeError(
                "%s: Expecting type %s, but got %s (attr=%s, %s)" % (self, typeval, value, item, self.keys[item]))

    def __init__(self, **args):
        self.init(args)

    def __setattr__(self, item, value):
        self.__check_type(item, value)
        self.__dict__[item] = value

    def check_integrity(self):
        for attr, value in list(self.__dict__.items()):
            self.__check_type(attr, value)
        return True



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import theme

x_base = 300
y_base = 300


def xscale(x):
    return x * theme.scale_factor + x_base


def yscale(y):
    return y * theme.scale_factor + y_base


def nscale(x):
    return x * theme.scale_factor


def nscale_seq(x):
    return list(map(nscale, x))



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
#
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import math
import re
from xml.dom.minidom import Document, Comment

from . import basecanvas
from . import version
from .scaling import *

# Note we flip all y-coords and negate all angles because SVG's coord
# system is inverted wrt postscript/PDF - note it's not enough to
# scale(1,-1) since that turns text into mirror writing with wrong origin

_comment_p = 0  # whether comment() writes output


def _svgcolor(color):
    """
    Convert a PyChart color object to an SVG rgb() value.
    See color.py.
    """
    return 'rgb(%d,%d,%d)' % tuple([int(255 * x) for x in [color.r, color.g, color.b]])


def _parse_style_str(s):
    """
    Take an SVG 'style' attribute string like 'stroke:none;fill:black'
    and parse it into a dictionary like {'stroke' : 'none', 'fill' : 'black'}.
    """
    styledict = {}
    if s:
        # parses L -> R so later keys overwrite earlier ones
        for keyval in s.split(';'):
            l = keyval.strip().split(':')
            if l and len(l) == 2: styledict[l[0].strip()] = l[1].strip()
    return styledict


def _make_style_str(styledict):
    """
    Make an SVG style string from the dictionary. See also _parse_style_str also.
    """
    s = ''
    for key in list(styledict.keys()):
        s += "%s:%s;" % (key, styledict[key])
    return s


def _protect_current_children(elt):
    # If elt is a group, check to see whether there are any non-comment
    # children, and if so, create a new group to hold attributes
    # to avoid affecting previous children.  Return either the current
    # elt or the newly generated group.
    if (elt.nodeName == 'g'):
        for kid in elt.childNodes:
            if kid.nodeType != Comment.nodeType:
                g = elt.ownerDocument.createElement('g')
                g.setAttribute('auto', '')
                if _comment_p:
                    g.appendChild(g.ownerDocument.createComment
                                  ('auto-generated group'))
                elt.appendChild(g)
                elt = g
                break
    return elt


class T(basecanvas.T):
    def __init__(self, fname):
        basecanvas.T.__init__(self)
        self.__out_fname = fname
        self.__xmin, self.__xmax, self.__ymin, self.__ymax = 0, 0, 0, 0
        self.__doc = Document()
        self.__doc.appendChild(self.__doc.createComment
                               ('Created by PyChart ' + version.version + ' ' + version.copyright))
        self.__svg = self.__doc.createElement('svg')  # the svg doc
        self.__doc.appendChild(self.__svg)
        self.__defs = self.__doc.createElement('defs')  # for clip paths
        self.__svg.appendChild(self.__defs)
        self.__cur_element = self.__svg
        self.gsave()  # create top-level group for dflt styles
        self.__update_style(font_family=theme.default_font_family,
                            font_size=theme.default_font_size,
                            font_style='normal',
                            font_weight='normal',
                            font_stretch='normal',
                            fill='none',
                            stroke='rgb(0,0,0)',  # SVG dflt none, PS dflt blk
                            stroke_width=theme.default_line_width,
                            stroke_linejoin='miter',
                            stroke_linecap='butt',
                            stroke_dasharray='none')

    def __update_style(self, **addstyledict):
        elt = _protect_current_children(self.__cur_element)

        # fetch the current styles for this node
        my_style_dict = _parse_style_str(elt.getAttribute('style'))

        # concat all ancestor style strings to get default styles for this node
        parent, s = elt.parentNode, ''
        while parent.nodeType != Document.nodeType:
            # prepend parent str so later keys will override earlier ones
            s = parent.getAttribute('style') + s
            parent = parent.parentNode
        default_style_dict = _parse_style_str(s)

        # Do some pre-processing on the caller-supplied add'l styles
        # Convert '_' to '-' so caller can specify style tags as python
        # variable names, eg. stroke_width => stroke-width.
        # Also convert all RHS values to strs
        for key in list(addstyledict.keys()):
            k = re.sub('_', '-', key)
            addstyledict[k] = str(addstyledict[key])  # all vals => strs
            if (k != key): del addstyledict[key]

        for k in list(addstyledict.keys()):
            if (k in my_style_dict or  # need to overwrite it
                    (k not in default_style_dict) or  # need to set it
                    default_style_dict[k] != addstyledict[k]):  # need to override it
                my_style_dict[k] = addstyledict[k]

        s = _make_style_str(my_style_dict)
        if s: elt.setAttribute('style', s)

        self.__cur_element = elt

    ####################################################################
    # methods below define the pychart backend device API

    # First are a set of methods to start, construct and finalize a path

    def newpath(self):  # Start a new path
        if (self.__cur_element.nodeName != 'g'):
            raise OverflowError("No containing group for newpath")
        # Just insert a new 'path' element into the document
        p = self.__doc.createElement('path')
        self.__cur_element.appendChild(p)
        self.__cur_element = p

    # This set of methods add data to an existing path element,
    # simply add to the 'd' (data) attribute of the path elt

    def moveto(self, x, y):  #
        if (self.__cur_element.nodeName != 'path'):
            raise OverflowError("No path for moveto")
        d = ' '.join([self.__cur_element.getAttribute('d'), 'M', repr(x), repr(-y)]).strip()
        self.__cur_element.setAttribute('d', d)

    def lineto(self, x, y):
        if (self.__cur_element.nodeName != 'path'):
            raise OverflowError("No path for lineto")
        d = ' '.join([self.__cur_element.getAttribute('d'), 'L', repr(x), repr(-y)]).strip()
        self.__cur_element.setAttribute('d', d)

    def path_arc(self, x, y, radius, ratio, start_angle, end_angle):
        # mimic PS 'arc' given radius, yr/xr (=eccentricity), start and
        # end angles.  PS arc draws from CP (if exists) to arc start,
        # then draws arc in counterclockwise dir from start to end
        # SVG provides an arc command that draws a segment of an
        # ellipse (but not a full circle) given these args:
        # A xr yr rotate majorArcFlag counterclockwiseFlag xe ye
        # We don't use rotate(=0) and flipped axes => all arcs are clockwise

        if (self.__cur_element.nodeName != 'path'):
            raise OverflowError("No path for path_arc")

        self.comment('x=%g, y=%g, r=%g, :=%g, %g-%g'
                     % (x, y, radius, ratio, start_angle, end_angle))

        xs = x + radius * math.cos(2 * math.pi / 360. * start_angle)
        ys = y + ratio * radius * math.sin(2 * math.pi / 360. * start_angle)
        xe = x + radius * math.cos(2 * math.pi / 360. * end_angle)
        ye = y + ratio * radius * math.sin(2 * math.pi / 360. * end_angle)
        if (end_angle < start_angle):  # make end bigger than start
            while end_angle <= start_angle:  # '<=' so 360->0 becomes 360->720
                end_angle += 360
        full_circ = (end_angle - start_angle >= 360)  # draw a full circle?

        d = self.__cur_element.getAttribute('d')
        d += ' %s %g %g' % (d and 'L' or 'M', xs, -ys)  # draw from CP, if exists
        if (radius > 0):  # skip, eg. 0-radius 'rounded' corners which blowup
            if (full_circ):
                # If we're drawing a full circle, move to the end coord
                # and draw half a circle to the reflected xe,ye
                d += ' M %g %g A %g %g 0 1 0 %g %g' % (xe, -ye,
                                                       radius, radius * ratio,
                                                       2 * x - xe, -(2 * y - ye))
            # Draw arc from the CP (either reflected xe,ye for full circle else
            # xs,ys) to the end coord - note with full_circ the
            # 'bigArcFlag' value is moot, with exactly 180deg left to draw
            d += ' A %g %g 0 %d 0 %g %g' % (radius, radius * ratio,
                                            end_angle - start_angle > 180,
                                            xe, -ye)
        self.__cur_element.setAttribute('d', d.strip())

    def curveto(self, x1, y1, x2, y2, x3, y3):
        # Equivalent of PostScript's x1 y1 x2 y2 x3 y3 curveto which
        # draws a cubic bezier curve from curr pt to x3,y3 with ctrl points
        # x1,y1, and x2,y2
        # In SVG this is just d='[M x0 y0] C x1 y1 x2 y2 x3 y3'
        # ! I can't find an example of this being used to test it
        if (self.__cur_element.nodeNode != 'path'):
            raise OverflowError("No path for curveto")
        d = ' '.join([self.__cur_element.getAttribute('d'), 'C',
                      repr(x1), repr(-y1), repr(x2), repr(-y2), repr(x3), repr(-y3), ]).strip()
        self.__cur_element.setAttribute('d', d)

    def closepath(self):  # close back to start of path
        if (self.__cur_element.nodeName != 'path'):
            raise OverflowError("No path for closepath")
        d = ' '.join([self.__cur_element.getAttribute('d'), 'Z']).strip()
        self.__cur_element.setAttribute('d', d)

    # Next we have three methods for finalizing a path element,
    # either fill it, clip to it, or draw it (stroke)
    # canvas.polygon() can generate fill/clip cmds with
    # no corresponding path so just ignore them
    def stroke(self):
        if (self.__cur_element.nodeName != 'path'):
            self.comment('No path - ignoring stroke')
            return
        self.__update_style(fill='none')
        self.__cur_element = self.__cur_element.parentNode

    def fill(self):
        if (self.__cur_element.nodeName != 'path'):
            self.comment('No path - ignoring fill')
            return
        self.__update_style(stroke='none')
        self.__cur_element = self.__cur_element.parentNode

    def clip_sub(self):
        if (self.__cur_element.nodeName != 'path'):
            self.comment('No path - ignoring clip')
            return

        # remove the current path from the tree ...
        p = self.__cur_element
        self.__cur_element = p.parentNode
        self.__cur_element.removeChild(p)

        # ... add it to a clipPath elt in the defs section
        clip = self.__doc.createElement('clipPath')
        clipid = 'clip' + repr(len(self.__defs.childNodes))
        clip.setAttribute('id', clipid)
        clip.appendChild(p)
        self.__defs.appendChild(clip)

        # ... update the local style to point to it
        self.__update_style(clip_path='url(#%s)' % clipid)

    # The text_xxx routines specify the start/end and contents of text
    def text_begin(self):
        if (self.__cur_element.nodeName != 'g'):
            raise ValueError("No group for text block")
        t = self.__doc.createElement('text')
        self.__cur_element.appendChild(t)
        self.__cur_element = t

    def text_moveto(self, x, y, angle):
        if (self.__cur_element.nodeName != 'text'):
            raise ValueError("No text for moveto")
        self.__cur_element.setAttribute('x', repr(x))
        self.__cur_element.setAttribute('y', repr(-y))
        if (angle):
            self.__cur_element.setAttribute('transform',
                                            'rotate(%g,%g,%g)' % (-angle, x, -y))

    def text_show(self, font_name, size, color, string):
        if (self.__cur_element.nodeName != 'text'):
            raise ValueError("No text for show")

        # PyChart constructs a postscript font name, for example:
        #
        # Helvetica Helvetica-Bold Helvetica-Oblique Helvetica-BoldOblique
        # Helvetica-Narrow Times-Roman Times-Italic
        # Symbol Palatino-Roman Bookman-Demi Courier AvantGarde-Book
        #
        # We need to deconstruct this to get the font-family (the
        # piece before the '-'), and other characteristics.
        # Note that 'Courier' seems to correspond to SVGs 'CourierNew'
        # and that the SVG Symbol font is Unicode where the ascii text
        # 'Symbol' doesn't create greek characters like 'Sigma ...' -
        # should really pass a unicode string, or provide translation
        #
        # SVG defines:
        # font-style = normal (aka roman) | italic | oblique
        # font-weight = normal | bold (aka demi?)
        # font-stretch = normal | wider | narrower | ultra-condensed |
        #          extra-condensed | condensed | semi-condensed |
        #          semi-expanded | expanded | extra-expanded | ultra-expanded
        # ('narrow' seems to correspond to 'condensed')
        font_name, modifiers = re.match(r'([^-]*)(-.*)?', font_name).groups()
        if font_name == 'Courier': font_name = 'CourierNew'
        font_style = font_weight = font_stretch = 'normal'
        if modifiers:
            if re.search('Italic', modifiers):
                font_style = 'italic'
            elif re.search('Oblique', modifiers):
                font_style = 'oblique'
            if re.search('Bold|Demi', modifiers): font_weight = 'bold'
            if re.search('Narrow', modifiers): font_stretch = 'condensed'
        # ! translate ascii symbol font chars -> unicode (see www.unicode.org)
        # ! http://www.unicode.org/Public/MAPPINGS/VENDORS/ADOBE/symbol.txt
        # ! but xml Text element writes unicode chars as '?' to XML file...
        string = re.sub(r'\\([()])', r'\1', string)  # unescape brackets
        self.__update_style(fill=_svgcolor(color),
                            stroke='none',
                            font_family=font_name,
                            font_size=size,
                            font_style=font_style,
                            font_weight=font_weight,
                            font_stretch=font_stretch)
        self.__cur_element.appendChild(self.__doc.createTextNode(string.encode('utf-8')))

    def text_end(self):
        if (self.__cur_element.nodeName != 'text'):
            raise ValueError("No text for close")
        self.__cur_element = self.__cur_element.parentNode

    # Three methods that change the local style of elements
    # If applied to a group, they persist until the next grestore,
    # If applied within a path element, they only affect that path -
    # although this may not in general correspond to (say) PostScript
    # behavior, it appears to correspond to reflect mode of use of this API
    def set_fill_color(self, color):
        self.__update_style(fill=_svgcolor(color))

    def set_stroke_color(self, color):
        self.__update_style(stroke=_svgcolor(color))

    def set_line_style(self, style):  # see line_style.py
        linecap = {0: 'butt', 1: 'round', 2: 'square'}
        linejoin = {0: 'miter', 1: 'round', 2: 'bevel'}
        if style.dash:
            dash = ','.join(map(str, style.dash))
        else:
            dash = 'none'
        self.__update_style(stroke_width=style.width,
                            stroke=_svgcolor(style.color),
                            stroke_linecap=linecap[style.cap_style],
                            stroke_linejoin=linejoin[style.join_style],
                            stroke_dasharray=dash)

    # gsave & grestore respectively push & pop a new context to hold
    # new style and transform parameters.  push/pop transformation are
    # similar but explicitly specify a coordinate transform at the
    # same time
    def gsave(self):
        if (self.__cur_element.nodeName not in ['g', 'svg']):
            raise ValueError("No group for gsave")
        g = self.__doc.createElement('g')
        self.__cur_element.appendChild(g)
        self.__cur_element = g

    def grestore(self):
        if (self.__cur_element.nodeName != 'g'):
            raise ValueError("No group for grestore")
        # first pop off any auto-generated groups (see protectCurrentChildren)
        while (self.__cur_element.hasAttribute('auto')):
            self.__cur_element.removeAttribute('auto')
            self.__cur_element = self.__cur_element.parentNode
        # then pop off the original caller-generated group
        self.__cur_element = self.__cur_element.parentNode

    def push_transformation(self, baseloc, scale, angle, in_text=0):
        # ? in_text arg appears to always be ignored

        # In some cases this gets called after newpath, with
        # corresonding pop_transformation called after the path is
        # finalized so we check specifically for that, and generate
        # an enclosing group to hold the incomplete path element
        # We could add the transform directly to the path element
        # (like we do with line-style etc) but that makes it harder
        # to handle the closing 'pop' and might lead to inconsitency
        # with PostScript if the closing pop doesn't come right after
        # the path element

        elt = self.__cur_element
        if elt.nodeName == 'g':
            elt = None
        elif (elt.nodeName == 'path' and not elt.hasAttribute('d')):
            g = elt.parentNode
            g.removeChild(elt)
            self.__cur_element = g
        else:
            raise ValueError("Illegal placement of push_transformation")

        t = ''
        if baseloc:
            t += 'translate(%g,%g) ' % (baseloc[0], -baseloc[1])
        if angle:
            t += 'rotate(%g) ' % -angle
        if scale:
            t += 'scale(%g,%g) ' % tuple(scale)

        self.gsave()
        self.__cur_element.setAttribute('transform', t.strip())
        if elt:  # elt has incomplete 'path' or None
            self.__cur_element.appendChild(elt)
            self.__cur_element = elt

    def pop_transformation(self, in_text=0):  # ? in_text unused?
        self.grestore()

    # If verbose, add comments to the output stream (helps debugging)
    def comment(self, string):
        if _comment_p:
            self.__cur_element.appendChild(self.__doc.createComment(string))

    # The verbatim method is currently not supported - presumably with
    # the SVG backend the user would require access to the DOM since
    # we're not directly outputting plain text here
    def verbatim(self, string):
        self.__cur_element.appendChild(self.__doc.createComment('verbatim not implemented: ' + string))

    # The close() method finalizes the SVG document and flattens the
    # DOM document to XML text to the specified file (or stdout)
    def close(self):
        basecanvas.T.close(self)
        self.grestore()  # matching the gsave in __init__
        if (self.__cur_element.nodeName != 'svg'):
            raise ValueError("Incomplete document at close!")

        # Don't bother to output an empty document - this can happen
        # when we get close()d immediately by theme reinit
        if (len(self.__svg.childNodes[-1].childNodes) == 0):
            return

        fp, need_close = self.open_output(self.__out_fname)
        bbox = theme.adjust_bounding_box([self.__xmin, self.__ymin,
                                          self.__xmax, self.__ymax])
        self.__svg.setAttribute('viewBox', '%g %g %g %g'
                                % (xscale(bbox[0]),
                                   -yscale(bbox[3]),
                                   xscale(bbox[2]) - xscale(bbox[0]),
                                   yscale(bbox[3]) - yscale(bbox[1])))
        self.__svg.setAttribute('xmlns', 'http://www.w3.org/2000/svg')
        self.__svg.setAttribute('xmlns:xlink', 'http://www.w3.org/1999/xlink')

        self.__doc.writexml(fp, '', '  ', '\n')
        if need_close:
            fp.close()



# -*- coding: utf-8 -*-
import os
import os.path
import re
import sys
import time
import traceback

import Purr
import Purr.Render
import Purr.RenderIndex
from Purr import dprint, dprintf, verbosity
from Purr.Render import quote_url


def _copy_update(sourcepath, destname):
    """Copy source to dest only if source is newer."""
    if sys.platform.startswith('linux'):
        return os.system("/bin/cp -ua '%s' '%s'" % (sourcepath, destname))
    else:
        return os.system("rsync -ua '%s' '%s'" % (sourcepath, destname))


def _move_update(sourcepath, destname):
    """Move source to dest only if source is newer."""
    if sys.platform.startswith('linux'):
        return os.system("/bin/mv -fu '%s' '%s'" % (sourcepath, destname))
    else:
        return os.system("rsync -ua --remove-source-files '%s' '%s'" % (sourcepath, destname))


class DataProduct(object):
    def __init__(self, filename=None, sourcepath=None, fullpath=None,
                 policy="copy", comment="",
                 timestamp=None, render=None,
                 quiet=False, archived=False):
        # This is the absolute pathname to the original data product
        self.sourcepath = Purr.canonizePath(sourcepath)
        # Base filename (w/o path) of data product within the log storage area.
        # Products may be renamed when they are moved or copied over to the log.
        self.filename = filename or (sourcepath and os.path.basename(sourcepath))
        # Full path to the DP within the log storage area.
        # This is None until a DP has been saved.
        self.fullpath = fullpath
        # Handling policy for DP: "copy","move","ignore", etc.
        self.policy = policy
        # Comment associated with DP
        self.comment = comment
        # Once a DP has been saved, this is the timestamp of data product at time of copy
        self.timestamp = timestamp
        # Name of renderer used to render this DP.
        self.render = render
        # if True, DP is watched quietly (i.e. Purr does not pop up windows on update)
        self.quiet = quiet
        # if True, DP has already been archived. This is False for new DPs until they're saved.
        self.archived = archived
        # if True, dp is ignored (policy is "ignore" or "banish")
        # not that policy should not be changed after a DP has been created
        self.ignored = policy in ("ignore", "banish")

    def set_policy(self, policy):
        if self.archived:
            raise TypeError("cannot change policy of archived DP")
        self.policy = policy
        self.ignored = policy in ("ignore", "banish")

    def subproduct_dir(self):
        """Returns name of subdirectory used to store subproducts of this data product
        during rendering."""
        return self.fullpath + ".purr-products"

    def remove_file(self):
        """Removes archived file associated with this DP"""
        if not self.fullpath or not self.archived:
            raise RuntimeError("""Can't remove a non-archived data product""")
        try:
            os.remove(self.fullpath)
        except:
            print(("Error removing %s: %s" % (self.fullpath, sys.exc_info()[1])))

    def remove_subproducts(self):
        """Removes all archived files subproducts associated with this DP"""
        if not self.fullpath or not self.archived:
            raise RuntimeError("""Can't remove a non-archived data product""")
        for root, dirs, files in os.walk(self.subproduct_dir(), topdown=False):
            for name in files:
                try:
                    os.remove(os.path.join(root, name))
                except:
                    pass
            for name in dirs:
                try:
                    os.remove(os.path.join(root, name))
                except:
                    pass

    def rename(self, newname):
        # rename file if needed
        if newname == self.filename:
            return None
        if not self.fullpath or not self.archived:
            raise RuntimeError("""Can't rename a non-archived data product""")
        dirname = os.path.dirname(self.fullpath)
        newpath = os.path.join(dirname, newname)
        oldsubname = self.subproduct_dir()
        try:
            os.rename(self.fullpath, newpath)
        except:
            print(("Error renaming %s to %s: %s" % (self.fullpath, newpath, sys.exc_info()[1])))
            return None
        # remove subproducts, if they exist -- need to re-render with new name anyway
        self.remove_subproducts()
        # set new paths and names
        self.fullpath = newpath
        self.filename = newname
        return newname

    def restore_from_archive(self, parent=None):
        """Function to restore a DP from archived copy
        Asks for confirmation along the way if parent is not None
        (in which case it will be the parent widget for confirmation dialogs)
        """
        from PyQt4.Qt import QMessageBox
        exists = os.path.exists(self.sourcepath)
        if parent:
            msg = """<P>Do you really want to restore <tt>%s</tt> from
            this entry's copy of <tt>%s</tt>?</P>""" % (self.sourcepath, self.filename)
            exists = os.path.exists(self.sourcepath)
            if exists:
                msg += """<P>Current file exists, and will be overwritten.</P>"""
                if QMessageBox.warning(parent, "Restoring from archive", msg,
                                       QMessageBox.Yes, QMessageBox.No) != QMessageBox.Yes:
                    return False
            else:
                if QMessageBox.question(parent, "Restoring from archive", msg,
                                        QMessageBox.Yes, QMessageBox.No) != QMessageBox.Yes:
                    return False
        busy = Purr.BusyIndicator()
        # remove file if in the way
        if exists:
            if os.system("/bin/rm -fr '%s'" % self.sourcepath):
                busy = None
                if parent:
                    QMessageBox.warning(parent, "Error removing file", """<P>
            There was an error removing %s. Archived copy was not restored.
            The text console may have more information.</P>""" % self.sourcepath,
                                        QMessageBox.Ok, 0)
                return False
        # unpack archived file
        if self.fullpath.endswith('.tgz'):
            parent_dir = os.path.dirname(self.sourcepath.rstrip('/'))
            os.system("/bin/rm -fr %s" % self.sourcepath)
            if os.system("tar zxf '%s' -C '%s'" % (self.fullpath, parent_dir)):
                busy = None
                if parent:
                    QMessageBox.warning(parent, "Error unpacking file", """<P>
            There was an error unpacking the archived version to %s. The text console may have more information.</P>""" % self.sourcepath,
                                        QMessageBox.Ok, 0)
                return False
        # else simply copy over
        else:
            if os.system("/bin/cp -a '%s' '%s'" % (self.fullpath, self.sourcepath)):
                busy = None
                if parent:
                    QMessageBox.warning(parent, "Error copying file", """<P>
            There was an error copying the archived version to %s. The text console may have more information.</P>""" % self.sourcepath,
                                        QMessageBox.Ok, 0)
                return False
        busy = None
        if parent:
            QMessageBox.information(parent, "Restored file", """<P>Restored %s from this entry's
          archived copy.</P>""" % self.sourcepath,
                                    QMessageBox.Ok, 0)
        return True


class LogEntry(object):
    """This represents a LogEntry object"""

    def __init__(self, timestamp=None, title="", comment="", dps=[], load=None, ignore=False):
        self.title, self.comment, self.dps = title, comment, dps
        self.timestamp = timestamp or int(time.time())
        self.pathname = None
        self.ignore = ignore
        # This is None for unsaved entries, and the pathname of the index.html file once
        # the entry is saved.
        # If the entry is edited, this is reset to None again.
        self.index_file = None
        # This file caches HTML code for inclusion into a top-level log index.
        self.cached_include = None
        # This is true if the cache is valid and up-to-date
        self.cached_include_valid = False
        # This is true if entry is updated and needs to be saved. Set to False in load(),
        # set to True when something changes, or if load() detects that renderers have been updated.
        self.updated = True
        if load:
            self.load(load)
        # QTreeWidgetItem associated with this log entry
        self.tw_item = None
        # next/prev/up links
        self._next_link = self._prev_link = self._up_link = None

    _entry_re = re.compile(".*/(entry|ignore)-(\d\d\d\d)(\d\d)(\d\d)-(\d\d)(\d\d)(\d\d)$")

    @staticmethod
    def isValidPathname(name):
        return os.path.isdir(name) and LogEntry._entry_re.match(name)

    def update(self, title=None, comment=None, dps=None, timestamp=None):
        self.updated = True
        if title is not None:
            self.title = title
        if comment is not None:
            self.comment = comment
        if dps is not None:
            self.dps = dps
        if timestamp is not None:
            self.timestamp = timestamp

    def load(self, pathname):
        """Loads entry from directory."""
        match = self._entry_re.match(pathname)
        if not match:
            return None
        self.ignore = (match.group(1) == "ignore")
        if not os.path.isdir(pathname):
            raise ValueError("%s: not a directory" % pathname)
        if not os.access(pathname, os.R_OK | os.W_OK):
            raise ValueError("%s: insufficient access privileges" % pathname)
        # parse index.html file
        parser = Purr.Parsers.LogEntryIndexParser(pathname)
        self.index_file = os.path.join(pathname, 'index.html')
        for i, line in enumerate(open(self.index_file)):
            try:
                parser.feed(line)
            except:
                dprintf(0, "parse error at line %d of %s\n", i, self.index_file)
                raise
        # set things up from parser
        try:
            self.timestamp = int(float(parser.timestamp))
        except:
            self.timestamp = int(time.time())
        self.title = getattr(parser, 'title', None)
        if self.title is None:
            self.title = "Malformed entry, probably needs to be deleted"
        self.comment = getattr(parser, 'comments', None) or ""
        self.dps = getattr(parser, 'dps', [])
        self.pathname = pathname
        # see if any data products have been removed on us
        self.dps = [dp for dp in self.dps if os.path.exists(dp.fullpath)]
        # see if the cached include file is up-to-date
        self.cached_include = cache = os.path.join(pathname, 'index.include.html')
        mtime = (os.path.exists(cache) or 0) and os.path.getmtime(cache)
        if mtime >= max(Purr.Render.youngest_renderer, os.path.getmtime(self.index_file)):
            dprintf(2, "entry %s has a valid include cache\n", pathname)
            self.cached_include_valid = True
        else:
            dprintf(2, "entry %s does not have a valid include cache\n", pathname)
            self.cached_include_valid = False
        # mark entry as unchanged, if renderers are older than index
        self.updated = (Purr.Render.youngest_renderer > os.path.getmtime(self.index_file))

    def _relIndexLink(self):
        """Returns relative link to index.html of this entry. Link will be of the form ../entry-xxx/index.html"""
        return os.path.join("..", os.path.basename(self.pathname), "index.html")

    def setLogDirectory(self, dirname):
        self.pathname = pathname = os.path.join(dirname,
                                                ((self.ignore and "ignore") or "entry") +
                                                time.strftime("-%Y%m%d-%H%M%S",
                                                              time.localtime(self.timestamp)))

    def save(self, dirname=None, refresh=0, refresh_index=True, emit_message=True):
        """Saves entry in the given directory. Data products will be copied over if not
        residing in that directory.
        'refresh' is a timestamp, passed to renderIndex(), causing all data products OLDER than the specified time to be regenerated.
        'refresh_index', if true, causes index files to be re-rendered unconditionally
        """
        if not refresh and not self.updated:
            return
        timestr = time.strftime("%Y%m%d-%H%M%S", time.localtime(self.timestamp))
        Purr.progressMessage("Rendering entry for %s" % timestr)
        if dirname:
            self.pathname = pathname = os.path.join(dirname, "%s-%s" %
                                                    (("ignore" if self.ignore else "entry"), timestr))
        elif not self.pathname:
            raise ValueError("Cannot save entry: pathname not specified")
        else:
            pathname = self.pathname
        # set timestamp
        if not self.timestamp:
            self.timestamp = int(time.time())
        # get canonized path to output directory
        pathname = Purr.canonizePath(pathname)
        if not os.path.exists(pathname):
            os.mkdir(pathname)
        # now save content
        # get device of pathname -- need to know whether we move or copy
        devnum = os.stat(pathname).st_dev
        # copy data products as needed
        dprintf(2, "saving entry %s, %d data products\n", pathname, len(self.dps))
        dps = []
        for dp in self.dps:
            # if archived, this indicates a previously saved data product, so ignore it
            # if ignored, no need to save the DP -- but keep it in list
            if dp.archived or dp.ignored:
                dprintf(3, "dp %s is archived or ignored, skipping\n", dp.sourcepath)
                dps.append(dp)
                continue
            # file missing for some reason (perhaps it got removed on us?) skip data product entirely
            if not os.path.exists(dp.sourcepath):
                dprintf(2, "data product %s missing, ignoring\n", dp.sourcepath)
                continue
            Purr.progressMessage("archiving %s" % dp.filename, sub=True)
            # get normalized source and destination paths
            dprintf(2, "data product: %s, rename %s, policy %s\n", dp.sourcepath, dp.filename, dp.policy)
            sourcepath = Purr.canonizePath(dp.sourcepath)
            destname = dp.fullpath = os.path.join(pathname, dp.filename)
            dprintf(2, "data product: %s -> %s\n", sourcepath, destname)
            # does the destination product already exist? skip if same file, else remove
            if os.path.exists(destname):
                if os.path.samefile(destname, sourcepath):
                    dprintf(2, "same file, skipping\n")
                    dp.timestamp = os.path.getmtime(destname)
                    dps.append(dp)
                    continue
                if os.system("/bin/rm -fr '%s'" % destname):
                    print(("Error removing %s, which is in the way of %s" % (destname, sourcepath)))
                    print("This data product is not saved.")
                    continue
            # for directories, compress with tar
            if os.path.isdir(sourcepath):
                sourcepath = sourcepath.rstrip('/')
                if dp.policy == "copy" or dp.policy.startswith("move"):
                    dprintf(2, "archiving to tgz\n")
                    if os.system("tar zcf '%s' -C '%s' '%s'" % (destname,
                                                                os.path.dirname(sourcepath),
                                                                os.path.basename(sourcepath))):
                        print(("Error archiving %s to %s" % (sourcepath, destname)))
                        print("This data product is not saved.")
                        continue
                    if dp.policy.startswith("move"):
                        os.system("/bin/rm -fr '%s'" % sourcepath)
            # else just a file
            else:
                # now copy/move it over
                if dp.policy == "copy":
                    dprintf(2, "copying\n")
                    if _copy_update(sourcepath, destname):
                        print(("Error copying %s to %s" % (sourcepath, destname)))
                        print("This data product is not saved.")
                        continue
                elif dp.policy.startswith('move'):
                    if _move_update(sourcepath, destname):
                        print(("Error moving %s to %s" % (sourcepath, destname)))
                        print("This data product is not saved.")
                        continue
            # success, set timestamp and append
            dp.timestamp = os.path.getmtime(destname)
            dp.archived = True
            dps.append(dp)
        # reset list of data products
        self.dps = dps
        # now write out content
        self.cached_include = os.path.join(pathname, 'index.include.html')
        self.cached_include_valid = False
        self.index_file = os.path.join(pathname, "index.html")
        self.generateIndex(refresh=refresh, refresh_index=refresh_index and time.time())
        self.updated = False

    def setPrevUpNextLinks(self, prev=None, up=None, next=None):
        """Sets Prev link to point to the LogEntry object "prev". Set that object's Next link to point to us. Sets the "up" link to the URL 'up' (if up != None.)
        Sets the Next link to the entry 'next' (if next != None), or to nothing if next == ''."""
        if prev is not None:
            if prev:
                self._prev_link = quote_url(prev._relIndexLink())
                prev._next_link = quote_url(self._relIndexLink())
            else:
                self._prev_link = None
        if up is not None:
            self._up_link = up and quote_url(up)
        if next is not None:
            self._next_link = next and quote_url(next._relIndexLink())

    def generateIndex(self, refresh=0, refresh_index=0):
        """Writes the index file"""
        open(self.index_file, "wt").write(self.renderIndex(refresh=refresh, refresh_index=refresh_index))

    def remove_directory(self):
        """Removes this entry's directory from disk"""
        if not self.pathname:
            return
        if os.system("/bin/rm -fr '%s'" % self.pathname):
            print("Error removing %s")

    def timeLabel(self):
        return time.strftime("%x %X", time.localtime(self.timestamp))

    def renderIndex(self, relpath="", refresh=0, refresh_index=0):
        """Returns HTML index code for this entry.
        If 'relpath' is empty, renders complete index.html file.
        If 'relpath' is not empty, then index is being included into a top-level log, and
        relpath should be passed to all sub-renderers.
        In this case the entry may make use of its cached_include file, if that is valid.
        If 'refresh' is set to a timestamp, then any subproducts (thumbnails, HTML caches, etc.) older than the timestamp will need to be regenerated.
        If 'refresh_index' is set to a timestamp, then any index files older than the timestamp will need to be regenerated.
        If 'relpath' is empty and 'prev', 'next' and/or 'up' is set, then Prev/Next/Up links will be inserted
        """
        # check if cache can be used
        refresh_index = max(refresh, refresh_index)
        dprintf(2, "%s: rendering HTML index with relpath='%s', refresh=%s refresh_index=%s\n", self.pathname, relpath,
                time.strftime("%x %X", time.localtime(refresh)),
                time.strftime("%x %X", time.localtime(refresh_index)))
        if relpath and self.cached_include_valid:
            try:
                if os.path.getmtime(self.cached_include) >= refresh_index:
                    dprintf(2, "using include cache %s\n", self.cached_include)
                    return open(self.cached_include).read()
                else:
                    dprintf(2, "include cache %s out of date, will regenerate\n", self.cached_include)
                    self.cached_include_valid = False
            except:
                print(("Error reading cached include code from %s, will regenerate" % self.cached_include))
                if verbosity.get_verbose() > 0:
                    dprint(1, "Error traceback follows:")
                    traceback.print_exc()
                self.cached_include_valid = False
        # form up attributes for % operator
        attrs = dict(self.__dict__)
        attrs['timestr'] = time.strftime("%x %X", time.localtime(self.timestamp))
        attrs['relpath'] = relpath
        html = ""
        # replace title and comments for ignored entries
        if self.ignore:
            attrs['title'] = "This is not a real log entry"
            attrs['comment'] = """This entry was saved by PURR because the user
      chose to ignore and/or banish some data products. PURR has stored this
      information here for its opwn internal and highly nefarious purposes.
      This entry is will not appear in the log."""
        # replace < and > in title and comments
        attrs['title'] = attrs['title'].replace("<", "&lt;").replace(">", "&gt;")
        # write header if asked
        if not relpath:
            icon = Purr.RenderIndex.renderIcon(24, "..")
            html += """<HTML><BODY>
      <TITLE>%(title)s</TITLE>""" % attrs
            if self._prev_link or self._next_link or self._up_link:
                html += """<DIV ALIGN=right><P>%s %s %s</P></DIV>""" % (
                    (self._prev_link and "<A HREF=\"%s\">&lt;&lt;Previous</A>" % self._prev_link) or "",
                    (self._up_link and "<A HREF=\"%s\">Up</A>" % self._up_link) or "",
                    (self._next_link and "<A HREF=\"%s\">Next&gt;&gt;</A>" % self._next_link) or ""
                )
            html += ("<H2>" + icon + """ <A CLASS="TITLE" TIMESTAMP=%(timestamp)d>%(title)s</A></H2>""") % attrs
        else:
            icon = Purr.RenderIndex.renderIcon(24)
            html += """
        <HR WIDTH=100%%>
        <H2>""" + icon + """ %(title)s</H2>""" % attrs
        # write comments
        html += """
        <DIV ALIGN=right><P><SMALL>Logged on %(timestr)s</SMALL></P></DIV>\n

        <A CLASS="COMMENTS">\n""" % attrs
        # add comments
        logmode = False
        for cmt in self.comment.split("\n"):
            cmt = cmt.replace("<", "&lt;").replace(">", "&gt;").replace("&lt;BR&gt;", "<BR>")
            html += """      <P>%s</P>\n""" % cmt
        html += """    </A>\n"""
        # add data products
        if self.dps:
            have_real_dps = bool([dp for dp in self.dps if not dp.ignored])
            if have_real_dps:
                html += """
        <H3>Data products</H3>
        <TABLE BORDER=1 FRAME=box RULES=all CELLPADDING=5>\n"""
            for dp in self.dps:
                dpattrs = dict(dp.__dict__)
                dpattrs['comment'] = dpattrs['comment'].replace("<", "&lt;"). \
                    replace(">", "&gt;").replace('"', "''")
                # if generating complete index, write empty anchor for each DP
                if not relpath:
                    if dp.ignored:
                        html += """
            <A CLASS="DP" SRC="%(sourcepath)s" POLICY="%(policy)s" COMMENT="%(comment)s"></A>\n""" % dpattrs
                    # write normal anchor for normal products
                    else:
                        dpattrs['relpath'] = relpath
                        dpattrs['basename'] = os.path.basename(dp.filename)
                        html += """
            <A CLASS="DP" FILENAME="%(filename)s" SRC="%(sourcepath)s" POLICY="%(policy)s" QUIET=%(quiet)d TIMESTAMP=%(timestamp).6f RENDER="%(render)s" COMMENT="%(comment)s"></A>\n""" % dpattrs
                # render a table row
                if not dp.ignored:
                    renderer = Purr.Render.makeRenderer(dp.render, dp, refresh=refresh)
                    html += Purr.Render.renderInTable(renderer, relpath)
            if have_real_dps:
                html += """
        </TABLE>"""
        # write footer
        if not relpath:
            html += "</BODY></HTML>\n"
        else:
            # now, write to include cache, if being included
            open(self.cached_include, 'w').write(html)
            self.cached_include_valid = True
        return html



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import os
import os.path
import sys
import tempfile

from . import basecanvas
from . import pscanvas
from .scaling import *


def _get_gs_path():
    """Guess where the Ghostscript executable is
    and return its absolute path name."""
    path = os.environ.get("PATH", os.defpath)

    for dir in path.split(os.pathsep):
        for name in ("gs", "gs.exe", "gswin32c.exe"):
            g = os.path.join(dir, name)
            if os.path.exists(g):
                return g
    raise Exception("Ghostscript not found. path=%s" % str(path))


class T(pscanvas.T):
    """This class is a special kind of canvas that runs ghostscript
    on the generated postscript contents. It is not used stand-alone, but as
    a component of PNG and X11 display functionality."""

    def __write_contents(self, fp):
        fp.write(pscanvas.preamble_text)
        for name, id in list(self.__font_ids.items()):
            fp.write("/%s {/%s findfont SF} def\n" % (id, name))
        fp.write("%d %d translate\n" % (-self.bbox[0], -self.bbox[1]))
        fp.writelines(self.__output_lines)
        fp.write("showpage end\n")
        fp.flush()

    def close(self):
        # Don't call pscanvas.T.close, as it creates a
        # ps file. 
        basecanvas.T.close(self)

    def start_gs(self, arg):
        self.bbox = theme.adjust_bounding_box([xscale(self.__xmin),
                                               yscale(self.__ymin),
                                               xscale(self.__xmax),
                                               yscale(self.__ymax)])

        gs_path = _get_gs_path()
        self.pipe_fp = None
        if self.__output_lines == []:
            return

        if sys.platform != "win32" and hasattr(os, "popen"):
            # UNIX-like systems
            cmdline = "\"%s\" -q %s -g%dx%d -q >/dev/null 2>&1" % \
                      (gs_path, arg,
                       self.bbox[2] - self.bbox[0],
                       self.bbox[3] - self.bbox[1])
            self.pipe_fp = os.popen(cmdline, "w")
            self.__write_contents(self.pipe_fp)
        else:
            # XXX should use mktemp, but need to support python<=2.2 as well.
            fname = tempfile.mktemp("xxx")
            fp = open(fname, "wb")
            self.__write_contents(fp)
            fp.close()
            cmdline = "\"%s\" -q %s -g%dx%d -q <%s >NUL" % \
                      (gs_path, arg,
                       self.bbox[2] - self.bbox[0],
                       self.bbox[3] - self.bbox[1], fname)
            os.system(cmdline)
            os.unlink(fname)

    def close_gs(self):
        if self.pipe_fp:
            self.pipe_fp.close()
            self.pipe_fp = None



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import bar_plot_doc
from . import chart_object
from . import error_bar
from . import fill_style
from . import legend
from . import line_style
from . import theme
from . import pychart_util
from .pychart_types import *

fill_styles = iter([])  # type: ignore

_keys = {
    "direction": (StringType, "vertical",
                  """The direction the growth of the bars. The value is either 'horizontal'
                  or 'vertical'."""),
    "data": (AnyType, None, pychart_util.data_desc),
    "data_label_offset": (CoordType, (0, 5),
                          "The location of data labels relative to the sample point. See also attribute data_label_format."),

    "data_label_format": (FormatType, None, """The
                          format string for the label displayed besides each
                          bar.  It can be a `printf' style format
                          string, or a two-parameter function that
                          takes (x,y) values and returns a string. """
                          + pychart_util.string_desc),

    "label": (StringType, "???", pychart_util.label_desc),
    "bcol": (IntType, 0,
             """Specifies the column from which base values (i.e., X values when attribute "direction" is "vertical", Y values otherwise) are extracted.
The
             combination of "data", "bcol", and "hcol" attributes defines
             the set of boxes drawn by this chart. See the
             below example:
             
@example
             d = [[5,10], [7,22], [8,25]]
             p = bar_plot.T(data = d, bcol = 1, hcol = 2)
@end example

             Here, three bars will be drawn. The X values of the bars
             will be 5, 7, and 8. The Y values of the bars will be
             10, 22, and 25, respectively. (In practice, because
             the values of bcol and hcol defaults to 1 and 2, you can
             write the above example just as "p = bar_plot.T(data = d)".
             """),
    "hcol": (IntType, 1,
             """The column from which the height of each bar is extracted.
             See also the description of the 'bcol' attribute."""),
    "line_style": (line_style.T, line_style.default,
                   "The style of the outer frame of each box."),
    "fill_style": (fill_style.T, lambda: next(fill_styles),
                   "Defines the fill style of each box.",
                   "The style is picked from standard styles round-robin."),
    "legend_line_style": (line_style.T, None,
                          """The line style used to draw a legend entry. Usually, the value is None, meaning that the value of "line_style" attribute is used."""),
    "legend_fill_style": (fill_style.T, None,
                          """The fill style used to draw a legend entry. Usually, the value is None, meaning that the value of "fill_style" attribute is used."""),

    "cluster": (TupleType, (0, 1), """This attribute is used to
    cluster multiple bar plots side by side in a single chart.
    The value should be a tuple of two integers. The second value should be equal to the total number of bar plots in the chart. The first value should be the relative position of this chart; 0 places this chart the leftmost, and N-1
    (where N is the 2nd value of this attribute) places this chart the rightmost. Consider the below example:

@example
    a = area.T(...)
    p1 = bar_plot.T(data = [[1,20][2,30]], cluster=(0,2))
    p2 = bar_plot.T(data = [[1,25],[2,10]], cluster=(1,2))
    a.add_plot(p1, p2)
    a.draw()
@end example

    In this example, one group of bars will be drawn side-by-side at
    position x=1, one with height 20, the other with height 25. The
    other two bars will be drawn side by side at position x=2, one
    with height 30 and the other with height 10.
    """),
    "width": (UnitType, 5, """Width of each box. 
@cindex width!- of bar plot
@cindex size!- of bar plot
"""),
    "cluster_sep": (UnitType, 0, """The separation between
    clustered boxes."""),
    "stack_on": (AnyType, None,
                 "The value must be either None or bar_plot.T. If not None, bars of this plot are stacked on top of another bar plot."),
    "error_minus_col": (IntType, -1,
                        """Specifies the column from which the depth of the errorbar is extracted.  This attribute is meaningful only when
                        error_bar != None.
                        """),
    "qerror_minus_col": (IntType, -1,
                         """The depth of the "quartile" errorbar is extracted from 
                         this column in data. This attribute is meaningful only
                         when error_bar != None. """),
    "error_plus_col": (IntType, -1,
                       """The depth of the errorbar is extracted from 
                       this column in data. This attribute is meaningful only
                       when error_bar != None."""),
    "qerror_plus_col": (IntType, -1,
                        """The depth of the "quartile" errorbar is extracted from 
                        this column in data. This attribute is meaningful only
                        when error_bar != None."""),
    "error_bar": (error_bar.T, None,
                  "Specifies the style of the error bar. <<error_bar>>"),
    "_abs_data": (ListType, None,
                  "Used only internally."),
}


def find_bar_plot(ar, nth):
    "Find the NTH barplot of the cluster in area AR."
    for plot in ar.plots():
        if isinstance(plot, T) and plot.cluster[0] == nth:
            return plot
    raise Exception("The %dth bar plot in the cluster not found." % nth)


class T(chart_object.T):
    __doc__ = bar_plot_doc.doc
    keys = _keys

    def check_integrity(self):
        chart_object.T.check_integrity(self)
        self.compute_abs_data()
        return True

    def compute_abs_data(self):
        if self._abs_data != None:
            return

        if self.stack_on == None:
            self._abs_data = self.data
        else:
            n = []
            for pair in self.data:
                self.stack_on.compute_abs_data()
                newpair = list(pair[:])
                newpair[self.hcol] = self.stack_on.get_value(newpair[self.bcol]) + pair[self.hcol]
                n.append(newpair)
            self._abs_data = n

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def get_value(self, bval):
        for pair in self._abs_data:
            if pair[self.bcol] == bval:
                return pair[self.hcol]
        raise ValueError(str(bval) + ": can't find the xval")

    def get_data_range(self, which):
        if self.direction == 'vertical':
            if which == 'X':
                return pychart_util.get_data_range(self._abs_data, self.bcol)
            else:
                return pychart_util.get_data_range(self._abs_data, self.hcol)
        else:
            assert self.direction == 'horizontal'
            if which == 'Y':
                return pychart_util.get_data_range(self._abs_data, self.bcol)
            else:
                return pychart_util.get_data_range(self._abs_data, self.hcol)

    def get_bar_width(self, ar, nth):
        off = 0
        for i in range(0, nth):
            plot = find_bar_plot(ar, i)
            off += plot.width + plot.cluster_sep
        return off

    def draw_vertical(self, ar, can):
        for pair in self.data:
            xval = pair[self.bcol]
            yval = pychart_util.get_sample_val(pair, self.hcol)

            if None in (xval, yval): continue

            ybot = 0
            if self.stack_on:
                ybot = self.stack_on.get_value(xval)
                yval += ybot

            totalWidth = self.get_bar_width(ar, self.cluster[1])
            firstX = ar.x_pos(xval) - totalWidth / 2.0
            thisX = firstX + self.get_bar_width(ar, self.cluster[0])

            can.rectangle(self.line_style, self.fill_style,
                          thisX, ar.y_pos(ybot), thisX + self.width,
                          ar.y_pos(yval))

            if self.error_bar:
                plus = pair[self.error_minus_col or self.error_plus_col]
                minus = pair[self.error_plus_col or self.error_minus_col]
                qplus = 0
                qminus = 0
                if self.qerror_minus_col or self.qerror_plus_col:
                    qplus = pair[self.qerror_minus_col or self.qerror_plus_col]
                    qminus = pair[self.qerror_plus_col or self.qerror_minus_col]
                if None not in (plus, minus, qplus, qminus):
                    self.error_bar.draw(can, (thisX + self.width / 2.0, ar.y_pos(yval)),
                                        ar.y_pos(yval - minus),
                                        ar.y_pos(yval + plus),
                                        ar.y_pos(yval - qminus),
                                        ar.y_pos(yval + qplus))

            if self.data_label_format:
                can.show(thisX + self.width / 2.0 + self.data_label_offset[0],
                         ar.y_pos(yval) + self.data_label_offset[1],
                         "/hC" + pychart_util.apply_format(self.data_label_format, (pair[self.bcol], pair[self.hcol]),
                                                           1))

    def draw_horizontal(self, ar, can):
        for pair in self.data:
            yval = pair[self.bcol]
            xval = pychart_util.get_sample_val(pair, self.hcol)

            if None in (xval, yval): continue

            xbot = 0
            if self.stack_on:
                xbot = self.stack_on.get_value(yval)
                xval += xbot
            totalWidth = self.get_bar_width(ar, self.cluster[1])
            firstY = ar.y_pos(yval) - totalWidth / 2.0
            thisY = firstY + self.get_bar_width(ar, self.cluster[0])

            can.rectangle(self.line_style, self.fill_style,
                          ar.x_pos(xbot), thisY,
                          ar.x_pos(xval), thisY + self.width)

            if self.data_label_format:
                can.show(ar.x_pos(xval) + self.data_label_offset[0],
                         thisY + self.width / 2.0 + self.data_label_offset[1],
                         "/vM/hL" + pychart_util.apply_format(self.data_label_format,
                                                              (pair[self.bcol], pair[self.hcol]), 1))

    def get_legend_entry(self):
        if self.label:
            return legend.Entry(line_style=(self.legend_line_style or self.line_style),
                                fill_style=(self.legend_fill_style or self.fill_style),
                                label=self.label)
        return None

    def draw(self, ar, can):
        assert self.check_integrity()
        can.clip(ar.loc[0], ar.loc[1],
                 ar.loc[0] + ar.size[0], ar.loc[1] + ar.size[1])

        if self.direction == "vertical":
            self.draw_vertical(ar, can)
        else:
            self.draw_horizontal(ar, can)

        can.endclip()


def init():
    global fill_styles
    fill_styles = fill_style.standards.iterate()


theme.add_reinitialization_hook(init)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import math
import sys


def inch_to_point(inch):
    return inch * 72.0


def point_to_inch(pt):
    return float(pt) / 72.0


def rotate(x, y, degree):
    """Rotate a coordinate around point (0,0).
    - x and y specify the coordinate.
    - degree is a number from 0 to 360.
    Returns a new coordinate.
    """
    radian = float(degree) * 2 * math.pi / 360.0
    newx = math.cos(radian) * x - math.sin(radian) * y
    newy = math.sin(radian) * x + math.cos(radian) * y
    return (newx, newy)


debug_level = 1


def warn(*strs):
    for s in strs:
        sys.stderr.write(str(s))
        sys.stderr.write(" ")
    sys.stderr.write("\n")


def info(*strs):
    if debug_level < 100:
        return
    for s in strs:
        sys.stderr.write(str(s))
    sys.stderr.write("\n")


def get_sample_val(l, col):
    if len(l) <= col:
        return None
    return l[col]


def get_data_list(data, col):
    # data = [ elem[col] for elem in data if elem[col] != None ]
    r = []
    for item in data:
        val = get_sample_val(item, col)
        if val != None:
            r.append(val)
    return r


def get_data_range(data, col):
    data = get_data_list(data, col)
    for item in data:
        if type(item) not in (int, int, float):
            print((type(item), item))
            raise TypeError("Non-number passed to data: %s" % (data))
    return (min(data), max(data))


def round_down(val, bound):
    return int(val / float(bound)) * bound


def round_up(val, bound):
    return (int((val - 1) / float(bound)) + 1) * bound


#
# Attribute type checking stuff
#

def new_list():
    return []


def union_dict(dict1, dict2):
    dict = dict1.copy()
    dict.update(dict2)
    return dict


def TextVAlignType(val):
    if val in ('T', 'B', 'M', None):
        return None
    return "Text vertical alignment must be one of T(op), B(ottom), or M(iddle).\n"


def TextAlignType(val):
    if val in ('C', 'R', 'L', None):
        return None
    return "Text horizontal alignment must be one of C(enter), R(ight), or L(eft)."


def apply_format(format, val, defaultidx):
    if format == None:
        return None
    elif type(format) == StringType:
        return format % val[defaultidx]
    else:
        return format(*val)


data_desc = "Specifies the data points. <<chart_data>>"
label_desc = "The label to be displayed in the legend. <<legend>>, <<font>>"
xcol_desc = """The column, within attribute "data", from which the X values of sample points are extracted. <<chart_data>>"""
ycol_desc = """The column, within attribute "data", from which the Y values of sample points are extracted. <<chart_data>>"""
tick_mark_desc = "Tick marks to be displayed at each sample point. <<tick_mark>>"
line_desc = "The style of the line. "


def interval_desc(w):
    return "When the value is a number, it specifies the interval at which %s are drawn. Otherwise, the value must be a function that takes two arguments, min and max X (or Y) values. It must return a list of numbers that specify the X or Y points at which %s are drawn." % (
    w, w)


shadow_desc = """The value is either None or a tuple. When non-None,
a drop-shadow is drawn beneath the object. X-off, and y-off specifies the
offset of the shadow relative to the object, and fill specifies the
style of the shadow (@pxref{module-fill-style})."""

string_desc = """The appearance of the string produced here can be
controlled using escape sequences. <<font>>"""


#
#

class symbol_lookup_table:
    def __init__(self, dict, objs):
        self.names = {}
        for name, val in list(dict.items()):
            for obj in objs.list():
                if val == obj:
                    self.names[val] = name
                    break

    def lookup(self, obj):
        if obj in self.names:
            return self.names[obj]
        return None



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import axis_doc
from . import chart_object
from . import font
from . import line_style
from . import theme
from . import pychart_util
from .pychart_types import *


class T(chart_object.T):
    keys = {
        "tic_interval": (IntervalType, None,
                         pychart_util.interval_desc("tick marks")),
        "tic_len": (UnitType, 6,
                    """The length of tick lines. The value can be negative, in which case the tick lines are drawn right of (or above) the axis."""),
        "minor_tic_interval": (IntervalType, None,
                               pychart_util.interval_desc("minor tick marks")),
        "minor_tic_len": (UnitType, 3,
                          """The length of minor tick marks.  The value can be negative, in which case the tick lines are drawn right of (or above) the axis."""),
        "line_style": (line_style.T, line_style.default,
                       "Specifies the style of axis and tick lines."),
        "label": (bytes, "axis label",
                  "The descriptive string displayed below (or to the left of) the axis. <<font>>."),
        "format": (FormatType, "%s",
                   """The format string for tick labels.
                   It can be a `printf' style format string, or 
                   a single-parameter function that takes an X (or Y) value
                   and returns a string. """ +
                   pychart_util.string_desc),
        "label_offset": (CoordType, (None, None),
                         """The location for drawing the axis label, 
                         relative to the middle point of the axis.
                         If the value is None, the label is displayed
                         below (or to the left of) of axis at the middle."""),
        "tic_label_offset": (CoordType, (0, 0),
                             """The location for drawing tick labels, 
                             relative to the tip of the tick line."""),
        "offset": (UnitType, 0,
                   """The location of the axis. 
                   The value of 0 draws the
                   axis at the left (for the Y axis) or bottom (for the X axis)
                   edge of the drawing area.
                   """)
    }


class X(T):
    keys = pychart_util.union_dict(T.keys,
                                   {"draw_tics_above": (IntType, 0,
                                                        "If true, tick lines and labels are drawn above the axis line.")})
    __doc__ = axis_doc.doc_x

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def draw_tics_and_labels(self, ar, can):
        assert self.check_integrity()
        y_base = ar.loc[1] + self.offset
        self.tic_interval = self.tic_interval or ar.x_grid_interval

        can.line(self.line_style, ar.loc[0], y_base,
                 ar.loc[0] + ar.size[0], y_base)

        tic_dic = {}
        max_tic_height = 0
        if self.draw_tics_above:
            sign = 1
        else:
            sign = -1

        for i in ar.x_tic_points(self.tic_interval):
            tic_dic[i] = 1
            ticx = ar.x_pos(i)

            string = "/hC" + pychart_util.apply_format(self.format, (i,), 0)

            (total_height, base_height) = font.text_height(string)
            max_tic_height = max(max_tic_height, total_height)
            if self.draw_tics_above:
                base_height = 0

            can.line(self.line_style, ticx, y_base,
                     ticx, y_base + sign * self.tic_len)
            can.show(ticx + self.tic_label_offset[0],
                     y_base + sign * (self.tic_len + base_height) + self.tic_label_offset[1],
                     string)

        if self.minor_tic_interval:
            for i in ar.x_tic_points(self.minor_tic_interval):
                if i in tic_dic:
                    # a major tic was drawn already.
                    pass
                else:
                    ticx = ar.x_pos(i)
                    can.line(self.line_style, ticx, y_base,
                             ticx, y_base + sign * self.minor_tic_len)

        self.draw_label(ar, can, (y_base + sign *
                                  (self.tic_len + max_tic_height + 10)))

    def draw_label(self, ar, can, ylabel):
        if self.label == None: return

        string = "/hC/vM" + self.label
        (label_height, base_height) = font.text_height(string)
        xlabel = ar.loc[0] + ar.size[0] / 2.0
        if self.label_offset[0] != None:
            xlabel += self.label_offset[0]
        if self.label_offset[1] != None:
            ylabel += self.label_offset[1]
        can.show(xlabel, ylabel, string)

    def draw(self, ar, can):
        assert self.check_integrity()
        self.tic_interval = self.tic_interval or ar.x_grid_interval
        y_base = ar.loc[1] + self.offset
        can.line(self.line_style, ar.loc[0], y_base,
                 ar.loc[0] + ar.size[0], y_base)
        self.draw_tics_and_labels(ar, can)


class Y(T):
    __doc__ = axis_doc.doc_y
    keys = pychart_util.union_dict(T.keys,
                                   {"draw_tics_right": (IntType, 0,
                                                        "If true, tick lines and labels are drawn right of the axis line.")})

    def draw_left(self, ar, can):
        x_base = ar.loc[0] + self.offset
        xmin = 999999
        tic_dic = {}
        for i in ar.y_tic_points(self.tic_interval):
            y_tic = ar.y_pos(i)
            tic_dic[i] = 1
            can.line(self.line_style, x_base, y_tic,
                     x_base - self.tic_len, y_tic)
            string = pychart_util.apply_format(self.format, (i,), 0)
            if self.tic_len > 0: string = "/hR" + string

            tic_height, base_height = font.text_height(string)
            x = x_base - self.tic_len + self.tic_label_offset[0]
            can.show(x, y_tic - tic_height / 2.0 + self.tic_label_offset[1],
                     string)
            xmin = min(xmin, x - font.text_width(string))

        if self.minor_tic_interval:
            for i in ar.y_tic_points(self.minor_tic_interval):
                if i in tic_dic:
                    # a major tic line was drawn already.
                    pass
                else:
                    y_tic = ar.y_pos(i)
                    can.line(self.line_style, x_base, y_tic,
                             x_base - self.minor_tic_len, y_tic)

        self.draw_label(ar, can, xmin - theme.default_font_size / 2.0)

    def draw_right(self, ar, can):
        x_base = ar.loc[0] + self.offset
        xmax = 0
        tic_dic = {}
        for i in ar.y_tic_points(self.tic_interval):
            y_tic = ar.y_pos(i)
            tic_dic[i] = 1
            can.line(self.line_style, x_base, y_tic,
                     x_base + self.tic_len, y_tic)
            string = pychart_util.apply_format(self.format, (i,), 0)
            if self.tic_len > 0: string = "/hL" + string

            tic_height, base_height = font.text_height(string)
            x = x_base + self.tic_len + self.tic_label_offset[0]
            can.show(x, y_tic - tic_height / 2.0 + self.tic_label_offset[1],
                     string)
            xmax = max(xmax, x + font.text_width(string))

        if self.minor_tic_interval:
            for i in ar.y_tic_points(self.minor_tic_interval):
                if i in tic_dic:
                    # a major tic line was drawn already.
                    pass
                else:
                    y_tic = ar.y_pos(i)
                    can.line(self.line_style, x_base, y_tic,
                             x_base + self.minor_tic_len, y_tic)

        self.draw_label(ar, can, xmax + theme.default_font_size)

    def draw_label(self, ar, can, xlabel):
        if self.label == None:
            return
        ylabel = ar.loc[1] + ar.size[1] / 2
        if self.label_offset[0] != None:
            xlabel += self.label_offset[0]
        if self.label_offset[1] != None:
            ylabel += self.label_offset[1]
        can.show(xlabel, ylabel, "/a90/hC" + self.label)

    def draw(self, ar, can):
        assert self.check_integrity()
        self.tic_interval = self.tic_interval or ar.y_grid_interval

        x_base = ar.loc[0] + self.offset
        can.line(self.line_style, x_base, ar.loc[1], x_base, ar.loc[1] + ar.size[1])
        if self.draw_tics_right:
            self.draw_right(ar, can)
        else:
            self.draw_left(ar, can)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import math

from . import coord


class T(coord.T):
    def get_canvas_pos(self, size, val, min, max):
        if val <= 0:
            return 0
        xminl = math.log(min)
        xmaxl = math.log(max)
        vl = math.log(val)
        return size * (vl - xminl) / float(xmaxl - xminl)

    def get_tics(self, min, max, interval):
        "Generate the list of places for drawing tick marks."
        v = []
        if min <= 0:
            raise Exception("Min value (%s) < 0 in a log coordinate." % min)
        x = min
        while x <= max:
            v.append(x)
            x = x * interval
        return v

    def get_min_max(self, dmin, dmax, interval):
        interval = interval or 10
        dmin = max(0, dmin)  # we can't have a negative value with a log scale.
        v = 1.0
        while v > dmin:
            v = v / interval
        dmin = v
        v = 1.0
        while v < dmax:
            v = v * interval
        dmax = v

        return dmin, dmax, interval



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#

def generate_doc(name, suffix="", append=0):
    if append:
        fp = open(name + "_doc.py", "a+")
    else:
        fp = open(name + "_doc.py", "w")
        fp.write("# automatically generated by generate_docs.py.\n")
    fp.write("doc" + suffix + "=\" \"\n")
    fp.close()


generate_doc("area")
generate_doc("arrow")
generate_doc("axis", "_x")
generate_doc("axis", "_y", 1)
generate_doc("bar_plot")
generate_doc("color")
generate_doc("error_bar", "_1")
generate_doc("error_bar", "_2", 1)
generate_doc("error_bar", "_3", 1)
generate_doc("error_bar", "_4", 1)
generate_doc("error_bar", "_5", 1)
generate_doc("error_bar", "_6", 1)
generate_doc("fill_style")
generate_doc("line_plot")
generate_doc("pie_plot")
generate_doc("text_box")
generate_doc("range_plot")
generate_doc("legend")
generate_doc("legend", "_entry", 1)
generate_doc("line_style")
generate_doc("tick_mark")



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import chart_object
from . import error_bar_doc
from . import fill_style
from . import line_style
from . import pychart_util
from . import object_set
from .pychart_types import *

__doc__ = """Pychart offers several styles of error bars. Some of them
only displays the min/max confidence interval, while others can display
quartiles in addition to min/max."""


class T(chart_object.T):
    keys = {}   # type: dict
    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    pass


# Two horizontal lines at min & max locations.
class error_bar1(T):
    __doc__ = error_bar_doc.doc_1
    keys = {"tic_len": (UnitType, 10, "Length of the horizontal bars"),
            "line_style": (line_style.T, line_style.default, "")
            }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def draw(self, can, loc, min, max, qmin=None, qmax=None):
        x = loc[0]
        y = min
        can.line(self.line_style, x - self.tic_len / 2.0, y, x + self.tic_len / 2.0, y)
        y = max
        can.line(self.line_style, x - self.tic_len / 2.0, y, x + self.tic_len / 2.0, y)


class error_bar2(T):
    __doc__ = error_bar_doc.doc_2
    keys = {"tic_len": (UnitType, 3,
                        "The length of the horizontal bars"),
            "hline_style": (line_style.T, line_style.default,
                            "The style of the horizontal bars."),
            "vline_style": (line_style.T, None,
                            "The style of the vertical bar.")
            }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def draw(self, can, loc, min, max, qmin=None, qmax=None):
        vline_style = self.vline_style
        if not vline_style:
            vline_style = self.hline_style
        x = loc[0]
        y1 = min
        can.line(self.hline_style, x - self.tic_len / 2.0, y1, x + self.tic_len / 2.0, y1)
        y2 = max
        can.line(self.hline_style, x - self.tic_len / 2.0, y2, x + self.tic_len / 2.0, y2)
        can.line(vline_style, x, y1, x, y2)


class error_bar3(T):
    # Tufte style
    __doc__ = "This style is endorsed by the Tufte's books. " \
              + error_bar_doc.doc_3
    keys = {"line_style": (line_style.T, line_style.default, "")
            }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def draw(self, can, loc, min, max, qmin, qmax):
        x = loc[0]
        can.line(self.line_style, x, min, x, qmin)
        can.line(self.line_style, x, qmax, x, max)


class error_bar4(T):
    __doc__ = error_bar_doc.doc_4
    keys = {"line_style": (line_style.T, line_style.default, ""),
            "fill_style": (fill_style.T, fill_style.gray70, ""),
            "box_width": (UnitType, 4, ""),
            "tic_len": (UnitType, 4, "")
            }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def draw(self, can, loc, min, max, qmin, qmax):
        x = loc[0]
        style = self.line_style
        y1 = min
        can.line(style, x - self.tic_len / 2.0, y1, x + self.tic_len / 2.0, y1)
        y2 = max
        can.line(style, x - self.tic_len / 2.0, y2, x + self.tic_len / 2.0, y2)
        can.line(style, x, y1, x, y2)

        can.rectangle(style, self.fill_style,
                      x - self.box_width / 2.0, qmin,
                      x + self.box_width / 2.0, qmax)


# vertical line
class error_bar5(T):
    __doc__ = error_bar_doc.doc_5
    keys = {"line_style": (line_style.T, line_style.default, "")
            }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def draw(self, can, loc, min, max, qmin=None, qmax=None):
        x = loc[0]
        y = loc[1]

        min = (min - y) * 1 + y
        max = (max - y) * 1 + y
        can.line(self.line_style, x, min, x, max)


# a box
class error_bar6(T):
    __doc__ = error_bar_doc.doc_6
    keys = {"line_style": (line_style.T, line_style.default, ""),
            "fill_style": (fill_style.T, fill_style.gray70, ""),
            "center_line_style": (line_style.T, line_style.T(width=0.5), ""),
            "box_width": (UnitType, 4, ""),
            }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def draw(self, can, loc, min, max, qmin=None, qmax=None):
        x = loc[0]
        y = loc[1]

        can.rectangle(self.line_style, self.fill_style,
                      x - self.box_width / 2.0, min,
                      x + self.box_width / 2.0, max)
        can.line(self.center_line_style,
                 x - self.box_width / 2.0, (min + max) / 2.0,
                 x + self.box_width / 2.0, (min + max) / 2.0)


bar1 = error_bar1()
bar2 = error_bar2()
bar3 = error_bar3()
bar4 = error_bar4()
bar5 = error_bar5()
bar6 = error_bar6()

standards = object_set.T(bar1, bar2, bar3, bar4, bar5, bar6)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
doc = """Attributes supported by this class are:
minor_tic_len(type:number):
    The length of minor tick marks. 
    default value=3.
tic_label_offset(type:(x,y)):
    The location where the tick labels is drawn. Relative to the 
    tip of the tick mark. 
    default value=(0, 0).
format(type:printf format string):
    The format string for tick labels. It can be a `printf' style 
    format string, or a single-parameter function that returns a 
    string. See also font. 
    default value=%s.
label_offset(type:(x,y) or None):
    The location where the axis label is drawn. Relative to the 
    left-bottom corner of the axis. 
    default value=(None, None).
grid_interval(type:Number or function):
    When the value is a number, it specifies the interval with which 
    grid lines are drawn. Otherwise, the value must be a function. 
    It must take no argument and return the list of numbers, which 
    specifies the X or Y points where grid lines are drawn. 
    default value=None.
tic_len(type:number):
    The length of tick lines 
    default value=6.
grid_style(type:line_style.T):
    
    default value=gray70dash3.
tic_interval(type:Number or function):
    When the value is a number, it specifies the interval with which 
    tick marks are drawn. Otherwise, the value must be a function. 
    It must take no argument and return the list of numbers, which 
    specifies the X or Y points where tick marks are drawn. 
    default value=None.
line_style(type:line_style.T):
    The style of tick lines. 
    default value=default.
label(type:str):
    The description of the axis. See also font. 
    default value=axis label.
minor_tic_interval(type:Number or function):
    When the value is a number, it specifies the interval with which 
    minor tick marks are drawn. Otherwise, the value must be a function. 
    It must take no argument and return the list of numbers, which 
    specifies the X or Y points where minor tick marks are drawn. 
    default value=None.
first_tic_value(type:number):
    The location of the first tick mark. Defaults to the x_range[0] 
    (or y_range[0]) of the area. 
    default value=None.
"""



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import copy

from . import fill_style
from . import line_style


def _draw_zap(can, p1, p2, style, pat):
    x = copy.deepcopy(p1)
    x.extend(p2)
    can.polygon(None, pat, x)
    can.lines(style, p1)
    can.lines(style, p2)


def zap_horizontally(can, style, pat, x1, y1, x2, y2, xsize, ysize):
    """Draw a horizontal "zapping" symbol on the canvas that shows
    that a graph is ripped in the middle.

    Parameter <fill_style> specifies the style for the zig-zag lines.
    PAT specifies the pattern with which the area is filled.
    The symbol is drawn in the rectangle (<x1>, <y1>) - (<x2>, <y2>).
    Each "zigzag" has the width <xsize>, height <ysize>."""

    assert isinstance(style, line_style.T)
    assert isinstance(pat, fill_style.T)

    points = []
    points2 = []
    x = x1
    y = y1
    while x < x2:
        points.append((x, y))
        points2.append((x, y + (y2 - y1)))
        x += xsize
        if y == y1:
            y += ysize
        else:
            y -= ysize

    points2.reverse()
    _draw_zap(can, points, points2, style, pat)


def zap_vertically(can, style, pat, x1, y1, x2, y2, xsize, ysize):
    """Draw a vertical "zapping" symbol on the canvas that shows
    that a graph is ripped in the middle.

    Parameter <fill_style> specifies the style for the zig-zag lines.
    PAT specifies the pattern with which the area is filled.
    The symbol is drawn in the rectangle (<x1>, <y1>) - (<x2>, <y2>).
    Each "zigzag" has the width <xsize>, height <ysize>."""

    points = []
    points2 = []
    x = x1
    y = y1
    while y < y2:
        points.append((x, y))
        points2.append((x + (x2 - x1), y))
        y += ysize
        if x == x1:
            x += xsize
        else:
            x -= xsize

    points2.reverse()
    _draw_zap(can, points, points2, style, pat)



# -*- coding: utf-8 -*-
# TODO:
# - add right-click options to New Entry menu: delete DP
# - store settings (such as pounce mode) inside the log
# - add a "Change dir" button. Integrate the catfight dialog into the
#   main window
# - in the browser: do not tell Purr to attach to a new directory
#   until the user presses the Purr button.
# - think of including verbatim code snippets in HTML

import configparser
import fcntl
import fnmatch
import glob
import os
import os.path
import re
import time
import traceback
from past.builtins import cmp

import Kittens.utils
from PyQt4.Qt import QObject, SIGNAL

import Purr
import Purr.Parsers
import Purr.Plugins
import Purr.Render
import Purr.RenderIndex
from Purr import Config, dprint, dprintf

# this string is used to create lock files
_lockstring = "%s:%d" % (os.uname()[1], os.getpid())


def parse_pattern_list(liststr):
    """Parses a list of filename patterns of the form "Description=ptt,patt,...;Description=patt,...
    Returns list of (description,[patterns]) tuples."""
    _re_patt = re.compile("^(.*)=(.*)")
    patterns = []
    for ss in liststr.split(';'):
        match = _re_patt.match(ss)
        if match:
            desc = match.group(1)
            patt = match.group(2).split(',')
            patterns.append((desc, patt))
    return patterns


def make_pattern_list(patterns):
    """Makes a string of filename patterns from a list of (description,[patterns]) tuples."""
    pattstrs = []
    for desc, patts in patterns:
        # ";," not allowed in descriptions or patterns
        patts = [p.replace(";", "").replace(",", "") for p in patts]
        desc = desc.replace(";", "").replace(",", "")
        pattstrs.append("%s=%s" % (desc, ','.join(patts)))
    return ";".join(pattstrs)


def _printexc(message, *args):
    print((message % args))
    traceback.print_exc()


def matches_patterns(filename, patterns):
    return bool([patt for patt in patterns if fnmatch.fnmatch(filename, patt)])


class Purrer(QObject):
    @staticmethod
    def is_purrlog(path):
        """Checks if path refers to a valid purrlog.
        Path must exist, and must contain either at least one directory called entry-YYYYMMDD-HHMMSS, or the file "dirconfig"
        """
        if not os.path.isdir(path):
            return False
        if list(filter(os.path.isdir, glob.glob(
                os.path.join(path, "entry-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]-[0-9][0-9][0-9][0-9][0-9][0-9]")))):
            return True
        return os.path.exists(os.path.join(path, "dirconfig"))

    class WatchedFile(QObject):
        """A WatchedFile represents a single file being watched for changes."""

        def __init__(self, path, quiet=None, mtime=None, survive_deletion=False):
            """Creates watched file at 'path'. The 'quiet' flag is simply stored.
            If 'mtime' is not None, this will be the file's last-changed timestamp.
            If 'mtime' is None, it will use os.path.getmtime().
            The survive_deletion flag is used to mark watchers that should stay active even if the underlying file
            disappears. Watchers for old data products are created with this flag.
            """
            QObject.__init__(self)
            self.path = path
            self.enabled = True
            self.quiet = quiet
            dprintf(3, "creating WatchedFile %s, mtime %s (%f)\n", self.path,
                    time.strftime("%x %X", time.localtime(mtime or self.getmtime())), mtime or 0)
            self.mtime = mtime or self.getmtime()
            self.survive_deletion = survive_deletion
            self.disappeared = False

        def getmtime(self):
            """Returns the file's modification time.
            Returns None on access error (i.e. file doesn't exist)"""
            try:
                return os.path.getmtime(self.path)
            except:
                return None

        def isUpdated(self):
            """Checks if file was updated (i.e. mtime changed) since last check. Returns True if so.
            Returns None on access error."""
            if not self.enabled:
                return None
            mtime = self.getmtime()
            if mtime is None:
                return None
            # clear disappeared flag
            self.disappeared = False
            # compare m,times -- add .1 sec margin since float numbers may get clobbered during
            # conversion
            updated = (mtime or 0) > (self.mtime or 0) + .1
            if updated:
                dprintf(4, "WatchedFile %s is updated: mtime %f %s, old mtime %f %s\n", self.path,
                        mtime, time.strftime("%x %X", time.localtime(mtime)),
                        self.mtime, time.strftime("%x %X", time.localtime(self.mtime)))
            self.mtime = mtime
            return updated

        def newFiles(self):
            """Checks if there are any new files, returns iterable of (full) paths.
            (For a single file, this is just the file itself, if it has been updated.)
            Returns None on access error."""
            updated = self.isUpdated()
            if updated is None:
                return None
            return (updated and [self.path]) or []

    class WatchedDir(WatchedFile):
        """A WatchedDir represents a directory being watched for new files.
        """

        def __init__(self, path, watch_patterns=[], ignore_patterns=[], **kw):
            """Initializes directory.
            'ignore_patterns' is a list of patterns to be ignored.
            'watch_patterns' is a list of patterns to be watched.
            New files will be reported only if they don't match any of the ignore patterns, or
            match a watch pattern.
            All other arguments as per WatchedFile
            """
            Purrer.WatchedFile.__init__(self, path, **kw)
            self.watch_patterns = watch_patterns
            self.ignore_patterns = ignore_patterns
            self._newfiles = []
            # the self.fileset attribute gives the current directory content
            try:
                self.fileset = set(os.listdir(self.path))
            except:
                _printexc("Error doing listdir(%s)" % self.path)
                traceback.print_exc()
                self.fileset = None;  # this indicates a read error
                return
            # check for files created after the supplied timestamp
            if self.getmtime() > self.mtime:
                dprintf(2, "%s modified since last run (%f vs %f), checking for new files\n", self.path,
                        self.getmtime(), self.mtime)
                for fname in self.fileset:
                    # ignore files from ignore list
                    if matches_patterns(fname, ignore_patterns) and not matches_patterns(fname, watch_patterns):
                        dprintf(5, "%s: matches ignore list but not watch list, skipping\n", fname)
                        continue
                    fullname = os.path.join(self.path, fname)
                    # check creation time against our timestamp
                    try:
                        ctime = os.path.getctime(fullname)
                    except:
                        _printexc("Error getting ctime for %s, ignoring", fname)
                        continue
                    if ctime > self.mtime:
                        dprintf(4, "%s: new file (created %s)\n", fullname,
                                time.strftime("%x %X", time.localtime(ctime)))
                        # append basename to _newfiles: full path added in newFiles() below
                        self._newfiles.append(fname)

        def newFiles(self):
            """Returns new files (since last call to newFiles, or since creation).
            Return value is an iterable of (full) paths.
            Returns None on access error."""
            if not self.enabled:
                return None
            if self.fileset is None:
                return None
            newfiles = set(self._newfiles);  # some newfiles may have been found in __init__
            self._newfiles = []
            updated = self.isUpdated()
            if updated is None:
                return None
            elif updated:
                try:
                    fileset1 = set(os.listdir(self.path))
                except:
                    _printexc("Error doing listdir(%s)" % self.path)
                    traceback.print_exc()
                    self.fileset = None
                    return None
                newfiles.update(fileset1.difference(self.fileset))
                self.fileset = fileset1
            # skip new files in ignore list
            # also skip new files with older timestamps -- these may have been restored from the archive
            nfs = []
            for file in newfiles:
                if matches_patterns(file, self.ignore_patterns) and not matches_patterns(file, self.watch_patterns):
                    continue
                path = os.path.join(self.path, file)
                # try:
                #  mtime = os.path.getmtime(path)
                # except:
                #  _printexc("Error doing getmtime(%s), omitting"%path)
                #  continue
                # if mtime >= self.mtime:
                ## Don't want to do the above check really: new files may actually have an mtime < dir mtime.
                ## If file is new, that's reason enough to include it.
                nfs.append(path)
            return nfs

    class WatchedSubdir(WatchedDir):
        """A WatchedSubdir represents a directory being watched for updates
        to specific files (called "canaries") within that directory. The directory itself
        is reported as a "new file" if the directory mtime changes, or a canary has changed.
        """

        def __init__(self, path, canary_patterns=[], **kw):
            Purrer.WatchedDir.__init__(self, path, **kw)
            self.canary_patterns = canary_patterns
            self.canaries = {}
            # if no read errors, make up list of canaries from canary patterns
            if self.fileset is not None:
                for fname in self.fileset:
                    if matches_patterns(fname, canary_patterns):
                        fullname = os.path.join(self.path, fname)
                        self.canaries[fullname] = Purrer.WatchedFile(fullname, mtime=self.mtime)
                        dprintf(3, "watching canary file %s, timestamp %s\n",
                                fullname, time.strftime("%x %X", time.localtime(self.mtime)))

        def newFiles(self):
            """Returns new files (since last call to newFiles, or since creation).
            The only possible new file is the subdirectory itself, which is considered
            new if updated, or if a canary has changed.
            Return value is an iterable, or None on access error."""
            # check directory itself for updates
            if self.fileset is None:
                return None
            # check for new files first
            newfiles = Purrer.WatchedDir.newFiles(self)
            if newfiles is None:
                return None
            # this timestamp is assigned to all canaries when directory has changed
            timestamp = time.time()
            canaries_updated = False
            # check for new canaries among new files
            if newfiles:
                dprintf(3, "directory %s is updated\n", self.path)
                for fname in newfiles:
                    if matches_patterns(os.path.basename(fname), self.canary_patterns):
                        self.canaries[fname] = Purrer.WatchedFile(fname, mtime=timestamp)
                        dprintf(3, "watching new canary file %s, timestamp %s\n",
                                fname, time.strftime("%x %X", time.localtime(timestamp)))
            # else check current canaries for updates
            else:
                for filename, watcher in list(self.canaries.items()):
                    updated = watcher.isUpdated()
                    if updated is None:
                        dprintf(2, "access error on canary %s, will no longer be watched", filename)
                        del self.canaries[filename]
                        continue
                    elif updated:
                        dprintf(3, "canary %s is updated\n", filename)
                        newfiles = True;  # this is treated as a bool below
                        break
                # now, if directory has updated, reset timestamps on all canaries
                if newfiles:
                    for watcher in list(self.canaries.values()):
                        watcher.mtime = timestamp
            # returns ourselves (as new file) if something has updated
            return (newfiles and [self.path]) or []

    def __init__(self, purrlog, watchdirs=None):
        QObject.__init__(self)
        # load and parse configuration
        # watched files
        watch = Config.get("watch-patterns", "Images=*fits,*FITS,*jpg,*png;TDL configuration=.tdl.conf")
        self._watch = parse_pattern_list(watch)
        self._watch_patterns = set()
        for desc, patts in self._watch:
            self._watch_patterns.update(patts)
        dprint(1, "watching patterns", self._watch_patterns)
        # quietly watched files (dialog is not popped up)
        watch = Config.get("watch-patterns-quiet", "TDL configuration=.tdl.conf")
        self._quiet = parse_pattern_list(watch)
        self._quiet_patterns = set()
        for desc, patts in self._quiet:
            self._quiet_patterns.update(patts)
        dprint(1, "quietly watching patterns", self._quiet_patterns)
        # ignored files
        ignore = Config.get("ignore-patterns",
                            "Hidden files=.*;Purr logs=*purrlog;MeqTree logs=meqtree.log;Python files=*.py*;Backup files=*~,*.bck;Measurement sets=*.MS,*.ms;CASA tables=table.f*,table.dat,table.info,table.lock")
        self._ignore = parse_pattern_list(ignore)
        self._ignore_patterns = set()
        for desc, patts in self._ignore:
            self._ignore_patterns.update(patts)
        dprint(1, "ignoring patterns", self._ignore_patterns)
        # watched subdirectories
        subdirs = Config.get("watch-subdirs", "MEP tables=*mep/funklets,table.dat")
        _re_patt = re.compile("^(.*)=(.*)/(.*)$")
        self._subdir_patterns = []
        for ss in subdirs.split(';'):
            match = _re_patt.match(ss)
            if match:
                desc = match.group(1)
                dir_patt = match.group(2).split(',')
                canary_patt = match.group(3).split(',')
                self._subdir_patterns.append((desc, dir_patt, canary_patt))
        dprint(1, "watching subdirectories", self._subdir_patterns)
        # attach to directories
        self.attached = False;  # will be True when we successfully attach
        self.other_lock = None;  # will be not None if another PURR holds a lock on this directory
        self.lockfile_fd = None
        self.lockfile_fobj = None
        self._attach(purrlog, watchdirs)

    def __del__(self):
        self.detach()

    def detach(self):
        if self.lockfile_fobj:
            try:
                self.lockfile_fobj.close()
                self.lockfile_fd = None
            except:
                pass
        if self.lockfile_fd:
            try:
                os.close(self.lockfile_fd)
                self.lockfile_fd = None
            except:
                pass
        self.attached = False

    class LockedError(RuntimeError):
        pass

    class LockFailError(RuntimeError):
        pass

    def _attach(self, purrlog, watchdirs=None):
        """Attaches Purr to a purrlog directory, and loads content.
        Returns False if nothing new has been loaded (because directory is the same),
        or True otherwise."""
        purrlog = os.path.abspath(purrlog)
        dprint(1, "attaching to purrlog", purrlog)
        self.logdir = purrlog
        self.indexfile = os.path.join(self.logdir, "index.html")
        self.logtitle = "Unnamed log"
        self.timestamp = self.last_scan_timestamp = time.time()
        self._initIndexDir()
        # reset internal state
        self.ignorelistfile = None
        self.autopounce = False
        self.watched_dirs = []
        self.entries = []
        self._default_dp_props = {}
        self.watchers = {}
        self.temp_watchers = {}
        self.attached = False
        self._watching_state = {}
        # check that we hold a lock on the directory
        self.lockfile = os.path.join(self.logdir, ".purrlock")
        # try to open lock file for r/w
        try:
            self.lockfile_fd = os.open(self.lockfile, os.O_RDWR | os.O_CREAT)
        except:
            raise Purrer.LockFailError("failed to open lock file %s for writing" % self.lockfile)
        # try to acquire lock on the lock file
        try:
            fcntl.lockf(self.lockfile_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except:
            other_lock = os.fdopen(self.lockfile_fd, 'r').read()
            self.lockfile_fd = None
            raise Purrer.LockedError(other_lock)
        # got lock, write our ID to the lock file
        global _lockstring
        try:
            self.lockfile_fobj = os.fdopen(self.lockfile_fd, 'w')
            self.lockfile_fobj.write(_lockstring)
            self.lockfile_fobj.flush()
            os.fsync(self.lockfile_fd)
        except:
            raise
        #      raise Purrer.LockFailError("cannot write to lock file %s"%self.lockfile)
        # load log state if log directory already exists
        if os.path.exists(self.logdir):
            _busy = Purr.BusyIndicator()
            if os.path.exists(self.indexfile):
                try:
                    parser = Purr.Parsers.LogIndexParser()
                    for line in open(self.indexfile):
                        parser.feed(line)
                    self.logtitle = parser.title or self.logtitle
                    self.timestamp = parser.timestamp or self.timestamp
                    dprintf(2, "attached log '%s', timestamp %s\n",
                            self.logtitle, time.strftime("%x %X", time.localtime(self.timestamp)))
                except:
                    traceback.print_exc()
                    print(("Error parsing %s, reverting to defaults" % self.indexfile))
            # load log entries
            entries = []
            for fname in os.listdir(self.logdir):
                pathname = os.path.join(self.logdir, fname)
                if Purr.LogEntry.isValidPathname(pathname):
                    try:
                        entry = Purr.LogEntry(load=pathname)
                        dprint(2, "loaded log entry", pathname)
                    except:
                        print(("Error loading entry %s, skipping" % fname))
                        traceback.print_exc()
                        continue
                    entries.append(entry)
                else:
                    dprint(2, fname, "is not a valid Purr entry")
            # sort log entires by timestamp
            import six
            if six.PY2:
                entries.sort(lambda a, b: cmp(a.timestamp, b.timestamp))
            else:
                from past.builtins import cmp
                from functools import cmp_to_key
                entries.sort(key = cmp_to_key(lambda a, b: cmp(a.timestamp, b.timestamp)))
            self.setLogEntries(entries, save=False)
            # update own timestamp
            if entries:
                self.timestamp = max(self.timestamp, entries[-1].timestamp)
        # else logfile doesn't exist, create it
        else:
            self._initIndexDir()
        # load configuration if it exists
        # init config file
        self.dirconfig = configparser.RawConfigParser()
        self.dirconfigfile = os.path.join(self.logdir, "dirconfig")
        if os.path.exists(self.dirconfigfile):
            try:
                self.dirconfig.read(self.dirconfigfile)
            except:
                print(("Error loading config file %s" % self.dirconfigfile))
                traceback.print_exc()
            # load directory configuration
            for dirname in self.dirconfig.sections():
                try:
                    watching = self.dirconfig.getint(dirname, "watching")
                except:
                    watching = Purr.WATCHED
                dirname = os.path.expanduser(dirname)
                self.addWatchedDirectory(dirname, watching, save_config=False)
        # start watching the specified directories
        for name in (watchdirs or []):
            self.addWatchedDirectory(name, watching=None)
        # Finally, go through list of ignored files and mark their watchers accordingly.
        # The ignorelist is a list of lines of the form "timestamp filename", giving the timestamp when a
        # file was last "ignored" by the purrlog user.
        self.ignorelistfile = os.path.join(self.logdir, "ignorelist")
        if os.path.exists(self.ignorelistfile):
            # read lines from file, ignore exceptions
            ignores = {}
            try:
                for line in open(self.ignorelistfile).readlines():
                    timestamp, policy, filename = line.strip().split(" ", 2)
                    # update dictiornary with latest timestamp
                    ignores[filename] = int(timestamp), policy
            except:
                print(("Error reading %s" % self.ignorelistfile))
                traceback.print_exc()
            # now scan all listed files, and make sure their watchers' mtime is no older than the given
            # last-ignore-timestamp. This ensures that we don't pounce on these files after restarting purr.
            for filename, (timestamp, policy) in list(ignores.items()):
                watcher = self.watchers.get(filename, None)
                if watcher:
                    watcher.mtime = max(watcher.mtime, timestamp)
        # init complete
        self.attached = True
        return True

    def setWatchedFilePatterns(self, watch, ignore=[]):
        self._watch = watch
        self._ignore = ignore
        self._watch_patterns = set()
        for desc, patts in self._watch:
            self._watch_patterns.update(patts)
        dprint(1, "watching patterns", self._watch_patterns)
        self._ignore_patterns = set()
        for desc, patts in self._ignore:
            self._ignore_patterns.update(patts)
        dprint(1, "ignoring patterns", self._ignore_patterns)
        Config.set("watch-patterns", make_pattern_list(self._watch))
        Config.set("ignore-patterns", make_pattern_list(self._ignore))

    def setWatchingState(self, path, watching, save_config=True):
        dprintf(2, "%s: watching state is %d\n", path, watching)
        self._watching_state[path] = watching
        path = Kittens.utils.collapseuser(path)
        if watching == Purr.REMOVED:
            if self.dirconfig.has_section(path):
                self.dirconfig.remove_section(path)
        else:
            if not self.dirconfig.has_section(path):
                self.dirconfig.add_section(path)
            self.dirconfig.set(path, "watching", str(watching))
        if save_config:
            self.dirconfig.write(open(self.dirconfigfile, 'wt'))

    def watchingState(self, pathname):
        return self._watching_state[pathname]

    def watchedDirectories(self):
        return [(dd, state) for dd, state in list(self._watching_state.items()) if state != Purr.REMOVED]

    def addWatchedDirectory(self, dirname, watching=Purr.WATCHED, save_config=True):
        """Starts watching the specified directories for changes"""
        # see if we're alredy watching this exact set of directories -- do nothing if so
        dirname = Purr.canonizePath(dirname)
        # do nothing if already watching
        if dirname in self.watched_dirs:
            dprint(1, "addWatchDirectory(): already watching %s\n", dirname)
            # watching=None means do not change the watch-state
            if watching is None:
                return
        else:
            if watching is None:
                watching = Purr.WATCHED
            # make watcher object
            wdir = Purrer.WatchedDir(dirname, mtime=self.timestamp,
                                     watch_patterns=self._watch_patterns, ignore_patterns=self._ignore_patterns)
            # fileset=None indicates error reading directory, so ignore it
            if wdir.fileset is None:
                print(("There was an error reading the directory %s, will stop watching it." % dirname))
                self.setWatchingState(dirname, Purr.REMOVED, save_config=True)
                return
            self.watchers[dirname] = wdir
            self.watched_dirs.append(dirname)
            dprintf(2, "watching directory %s, mtime %s, %d files\n",
                    dirname, time.strftime("%x %X", time.localtime(wdir.mtime)), len(wdir.fileset))
            # find files in this directory matching the watch_patterns, and watch them for changes
            watchset = set()
            for patt in self._watch_patterns:
                watchset.update(fnmatch.filter(wdir.fileset, patt))
            for fname in watchset:
                quiet = matches_patterns(fname, self._quiet_patterns)
                fullname = Purr.canonizePath(os.path.join(dirname, fname))
                if fullname not in self.watchers:
                    wfile = Purrer.WatchedFile(fullname, quiet=quiet, mtime=self.timestamp)
                    self.watchers[fullname] = wfile
                    dprintf(3, "watching file %s, timestamp %s, quiet %d\n",
                            fullname, time.strftime("%x %X", time.localtime(wfile.mtime)), quiet)
            # find subdirectories  matching the subdir_patterns, and watch them for changes
            for fname in wdir.fileset:
                fullname = Purr.canonizePath(os.path.join(dirname, fname))
                if os.path.isdir(fullname):
                    for desc, dir_patts, canary_patts in self._subdir_patterns:
                        if matches_patterns(fname, dir_patts):
                            quiet = matches_patterns(fname, self._quiet_patterns)
                            wdir = Purrer.WatchedSubdir(fullname, canary_patterns=canary_patts, quiet=quiet,
                                                        mtime=self.timestamp)
                            self.watchers[fullname] = wdir
                            dprintf(3, "watching subdirectory %s/{%s}, timestamp %s, quiet %d\n",
                                    fullname, ",".join(canary_patts),
                                    time.strftime("%x %X", time.localtime(wdir.mtime)), quiet)
                            break
        # set state and save config
        self.setWatchingState(dirname, watching, save_config=save_config)

    def setLogTitle(self, title, save=True):
        self.logtitle = title
        if save:
            self.save()

    def addLogEntry(self, entry, save=True):
        """This is called when a new log entry is created"""
        # create log directory if it doesn't exist
        # error will be thrown if this is not possible
        _busy = Purr.BusyIndicator()
        self._initIndexDir()
        # discard temporary watchers -- these are only used to keep track of
        # deleted files
        self.temp_watchers = {}
        # ignored entries are only there to carry info on ignored data products
        # All we do is save them, and update DP policies based on them
        if entry.ignore:
            entry.save(self.logdir)
        # proper entries are added to list
        else:
            self.entries.append(entry)
            Purr.progressMessage("Saving new log entry")
            # find previous entry -- skip over "ignore" entries
            for prev in self.entries[-2::-1]:
                if not prev.ignore:
                    break
            else:
                prev = None
            entry.setLogDirectory(self.logdir)
            entry.setPrevUpNextLinks(prev=prev, up=os.path.join("..", Purr.RenderIndex.INDEX))
            entry.save()
            self.timestamp = self.last_scan_timestamp
            # regenerate links of previous entry
            prev and prev.generateIndex()
            # and our log may need to be regenerated
            if save:
                self.save()
        self.updatePoliciesFromEntry(entry, new=True)

    def getLogEntries(self):
        return self.entries

    def setLogEntries(self, entries, save=True, update_policies=True):
        """Sets list of log entries. If save=True, saves the log. If update_policies=True, also updates default policies based
        on these entries"""
        prev = None;  # "previous" valid entry for "Prev" link
        uplink = os.path.join("..", Purr.RenderIndex.INDEX);  # "Up" link
        self.entries = []
        for entry in entries:
            if not entry.ignore:
                entry.setPrevUpNextLinks(prev=prev, up=uplink)
                prev = entry
                self.entries.append(entry)
        if save:
            self.save()
        if update_policies:
            # populate default policies and renames based on entry list
            self._default_dp_props = {}
            for entry in entries:
                self.updatePoliciesFromEntry(entry)
            dprint(4, "default policies:", self._default_dp_props)

    def deleteLogEntries(self, del_entries, save=True):
        """Deletes given list of entries from the log. Some entries' index.html files may be updated to reflect changes in prev/next links. If save=True, re-saves the log itself."""
        prev_valid = prev_deleted = None
        # fix Next/Prev links in remaining entries
        newents = []
        updated = set()
        for ent in self.entries:
            if not ent.ignore:
                if ent in del_entries:
                    prev_deleted = ent
                else:
                    if prev_deleted:
                        ent.setPrevUpNextLinks(prev=prev_valid)
                        prev_valid and prev_valid.generateIndex()
                        ent.generateIndex()
                        prev_deleted = None
                    prev_valid = ent
                    newents.append(ent)
            else:
                newents.append(ent)
        # fix last "Next" link if we had a deleted entry followed by nothing
        if prev_deleted and prev_valid:
            prev_valid.setPrevUpNextLinks(next='')
            prev_valid.generateIndex()
        # regenerate indices
        for ent in newents:
            if id(ent) in updated:
                ent.generateIndex()
        self.entries = newents
        if save:
            self.save()

    def updatePoliciesFromEntry(self, entry, new=True):
        # populate default policies and renames based on entry list
        for dp in entry.dps:
            # add default policy
            basename = os.path.basename(dp.sourcepath)
            self._default_dp_props[basename] = dp.policy, dp.filename, dp.comment
            dprintf(4, "file %s: default policy is %s\n", basename, dp.policy)
            # make new watchers for non-ignored files
            if dp.ignored:
                # if ignorelistfile is not set, then we're being called from within
                # _attach(), when older log entries are being loaded. No need to write the file then.
                if new and self.ignorelistfile and os.path.exists(dp.sourcepath):
                    try:
                        open(self.ignorelistfile, 'a').write(
                            "%d %s %s\n" % (os.path.getmtime(dp.sourcepath), dp.policy, dp.sourcepath))
                    except:
                        print(("Error writing %s" % self.ignorelistfile))
                        traceback.print_exc()
            else:
                watcher = self.watchers.get(dp.sourcepath, None)
                # if watcher already exists, update timestamp
                if watcher:
                    if watcher.mtime < dp.timestamp:
                        watcher.mtime = dp.timestamp
                        dprintf(4, "file %s, updating timestamp to %s\n",
                                dp.sourcepath, time.strftime("%x %X", time.localtime(dp.timestamp)))
                # else create new watcher
                else:
                    wfile = Purrer.WatchedFile(dp.sourcepath, quiet=dp.quiet, mtime=dp.timestamp, survive_deletion=True)
                    self.watchers[dp.sourcepath] = wfile
                    dprintf(4, "watching file %s, timestamp %s\n",
                            dp.sourcepath, time.strftime("%x %X", time.localtime(dp.timestamp)))

    def _relIndexLink(self):
        """Returns relative link to index.html of this entry. Link will be of the form ../entry-xxx/index.html"""
        return os.path.join("..", os.path.basename(self.pathname), "index.html")

    def _initIndexDir(self):
        """makes sure purrlog directory is properly set up"""
        if not os.path.exists(self.logdir):
            os.mkdir(self.logdir)
            dprint(1, "created", self.logdir)
        Purr.RenderIndex.initIndexDir(self.logdir)

    def save(self, refresh=False):
        """Saves the log.
        If refresh is set to a timestamp, will regenerate everything from scratch.
        """
        # create directory if it doesn't exist
        # error will be thrown if this is not possible
        _busy = Purr.BusyIndicator()
        Purr.progressMessage("Generating index in %s" % self.logdir)
        self._initIndexDir()
        # if refresh is True, re-save all entries.
        if refresh:
            refresh = time.time()
            for i, entry in enumerate(self.entries):
                entry.save(refresh=refresh)
        Purr.RenderIndex.writeLogIndex(self.logdir, self.logtitle, self.timestamp, self.entries, refresh=refresh)
        Purr.progressMessage("Wrote %s" % self.logdir)

    def rescan(self):
        """Checks files and directories on watchlist for updates, rescans them for new data products.
        If any are found, returns them. Skips those in directories whose watchingState is set to Purr.UNWATCHED.
        """
        if not self.attached:
            return
        dprint(5, "starting rescan")
        newstuff = {};  # this accumulates names of new or changed files. Keys are paths, values are 'quiet' flag.
        # store timestamp of scan
        self.last_scan_timestamp = time.time()
        # go through watched files/directories, check for mtime changes
        for path, watcher in list(self.watchers.items()):
            # get list of new files from watcher
            newfiles = watcher.newFiles()
            # None indicates access error, so drop it from watcher set
            if newfiles is None:
                if watcher.survive_deletion:
                    dprintf(5, "access error on %s, but will still be watched\n", watcher.path)
                else:
                    dprintf(2, "access error on %s, will no longer be watched\n", watcher.path)
                    del self.watchers[path]
                if not watcher.disappeared:
                    self.emit(SIGNAL("disappearedFile"), path)
                    watcher.disappeared = True
                continue
            dprintf(5, "%s: %d new file(s)\n", watcher.path, len(newfiles))
            # if a file has its own watcher, and is independently reported by a directory watcher, skip the directory's
            # version and let the file's watcher report it. Reason for this is that the file watcher may have a more
            # up-to-date timestamp, so we trust it over the dir watcher.
            newfiles = [p for p in newfiles if p is path or p not in self.watchers]
            # skip files in self._unwatched_paths
            newfiles = [filename for filename in newfiles if
                        self._watching_state.get(os.path.dirname(filename)) > Purr.UNWATCHED]
            # Now go through files and add them to the newstuff dict
            for newfile in newfiles:
                # if quiet flag is explicitly set on watcher, enforce it
                # if not pouncing on directory, also add quietly
                if watcher.quiet or self._watching_state.get(os.path.dirname(newfile)) < Purr.POUNCE:
                    quiet = True
                # else add quietly if file is not in the quiet patterns
                else:
                    quiet = matches_patterns(os.path.basename(newfile), self._quiet_patterns)
                # add file to list of new products. Since a file may be reported by multiple
                # watchers, make the quiet flag a logical AND of all the quiet flags (i.e. DP will be
                # marked as quiet only if all watchers report it as quiet).
                newstuff[newfile] = quiet and newstuff.get(newfile, True)
                dprintf(4, "%s: new data product, quiet=%d (watcher quiet: %s)\n", newfile, quiet, watcher.quiet)
                # add a watcher for this file to the temp_watchers list. this is used below
                # to detect renamed and deleted files
                self.temp_watchers[newfile] = Purrer.WatchedFile(newfile)
        # now, go through temp_watchers to see if any newly pounced-on files have disappeared
        for path, watcher in list(self.temp_watchers.items()):
            # get list of new files from watcher
            if watcher.newFiles() is None:
                dprintf(2, "access error on %s, marking as disappeared", watcher.path)
                del self.temp_watchers[path]
                self.emit(SIGNAL("disappearedFile"), path)
        # if we have new data products, send them to the main window
        return self.makeDataProducts(iter(list(newstuff.items())))

    def makeDataProducts(self, files, unbanish=False, unignore=False):
        """makes a list of DPs from a list of (filename,quiet) pairs.
        If unbanish is False, DPs with a default "banish" policy will be skipped.
        Symlinks will be resolved, and non-unique filenames removed from list.
        """
        paths = set()
        dps = []
        for filename, quiet in files:
            filename = filename.rstrip('/')
            sourcepath = Purr.canonizePath(filename)
            if sourcepath not in paths:
                paths.add(sourcepath)
                filename = os.path.basename(filename)
                policy, filename, comment = self._default_dp_props.get(filename, ("copy", filename, ""))
                dprintf(4, "%s: default policy is %s,%s,%s\n", sourcepath, policy, filename, comment)
                if policy == "banish":
                    if unbanish:
                        policy = "copy"
                    else:
                        continue
                if unignore and policy == "ignore":
                    policy = "copy"
                dps.append(Purr.DataProduct(filename=filename, sourcepath=sourcepath,
                                            policy=policy, comment=comment, quiet=quiet))
        import six
        from past.builtins import cmp
        from functools import cmp_to_key
        if six.PY3:
            return sorted(dps, key=cmp_to_key(lambda a, b: cmp(a.filename, b.filename)))
        else:
            return sorted(dps, lambda a, b: cmp(a.filename, b.filename))



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import arrow
from . import chart_object
from . import fill_style
from . import font
from . import legend
from . import pychart_util
from . import line_style
from . import pie_plot_doc
from . import text_box
from . import theme
from .pychart_types import *


class T(chart_object.T):
    __doc__ = pie_plot_doc.doc
    keys = {
        "start_angle": (NumberType, 90,
                        """The angle at which the first item is drawn."""),
        "center": (CoordType, None, "The location of the center of the pie."),
        "radius": (UnitType, None, "The radius of the pie."),
        "line_style": (line_style.T, line_style.default, "The style of the outer edge of each pie slice."),

        "fill_styles": (ListType, fill_style.standards.list(),  # type: ignore
                        """The fill style of each item. The length of the
                        list should be equal to the length of the data. 
                        """),
        "arc_offsets": (ListType, None,
                        """You can draw each pie "slice" shifted off-center.
                        This attribute, if non-None,
                        must be a number sequence whose length is equal to
                        the number of pie slices. The Nth value in arc_offsets
                        specify the amount of offset
                        (from the center of the circle)
                        for the Nth slice.
                        The value of None will draw all the slices
                        anchored at the center. 
                        """
                        ),
        "data": (AnyType, None, pychart_util.data_desc),
        "label_format": (FormatType, "%s",
                         "Format string of the label"),
        "label_col": (IntType, 0,
                      """The column, within "data", from which the labels of items are retrieved."""),
        "data_col": (IntType, 1,
                     """ The column, within "data", from which the data values are retrieved."""),
        "label_offset": (UnitType, None, "The distance from the center of each label."),
        "arrow_style": (arrow.T, None,
                        """The style of arrow that connects a label
                        to the corresponding "pie"."""),
        "label_line_style": (line_style.T, None, "The style of the frame surrounding each label."),
        "label_fill_style": (fill_style.T, fill_style.default, "The fill style of the frame surrounding each label."),
        "shadow": (ShadowType, None, pychart_util.shadow_desc)
    }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def _total(self):
        v = 0
        for val in self.data:
            v += val[self.data_col]
        return v

    def check_integrity(self):
        chart_object.T.check_integrity(self)

    def get_data_range(self, which):
        return (0, 1)

    def get_legend_entry(self):
        legends = []
        i = 0
        for val in self.data:
            fill = self.fill_styles[i]
            i = (i + 1) % len(self.fill_styles)
            legends.append(legend.Entry(line_style=self.line_style,
                                        fill_style=fill,
                                        label=val[self.label_col]))
        return legends

    def draw(self, ar, can):
        center = self.center
        if not center:
            center = (ar.loc[0] + ar.size[0] / 2.0,
                      ar.loc[1] + ar.size[1] / 2.0)
        radius = self.radius
        if not radius:
            radius = min(ar.size[0] / 2.0, ar.size[1] / 2.0) * 0.5

        label_offset = radius + (self.label_offset or radius * 0.1)

        total = self._total()
        i = 0
        cur_angle = self.start_angle
        for val in self.data:
            fill = self.fill_styles[i]
            degree = 360 * float(val[self.data_col]) / float(total)

            off = (0, 0)
            if len(self.arc_offsets) > i:
                off = pychart_util.rotate(self.arc_offsets[i], 0, cur_angle - degree / 2.0)
            x_center = center[0] + off[0]
            y_center = center[1] + off[1]

            can.ellipsis(self.line_style, fill,
                         x_center, y_center, radius, 1,
                         cur_angle - degree, cur_angle,
                         self.shadow)

            label = pychart_util.apply_format(self.label_format, val,
                                              self.label_col)
            if label != None:
                (x_label, y_label) = pychart_util.rotate(label_offset, 0, cur_angle - degree / 2.0)
                (x_arrowtip, y_arrowtip) = pychart_util.rotate(radius, 0, cur_angle - degree / 2.0)
                # Labels on left side of pie need
                # their text to avoid obscuring the pie
                if x_label < 0:
                    x_label = x_label - font.text_width(label)

                t = text_box.T(loc=(x_label + x_center, y_label + y_center),
                               text=label,
                               line_style=self.label_line_style,
                               fill_style=self.label_fill_style)
                if self.arrow_style:
                    t.add_arrow((x_arrowtip + x_center, y_arrowtip + y_center),
                                None, self.arrow_style)

                t.draw(can)
            cur_angle = (cur_angle - degree) % 360
            i = (i + 1) % len(self.fill_styles)


def init():
    old_val = T.keys["fill_styles"]
    T.keys["fill_styles"] = (old_val[0], fill_style.standards.list(),
                             old_val[2])


theme.add_reinitialization_hook(init)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
version = "1.39"
copyright = "Copyright 1999-2006 Yasushi Saito"



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
class T(object):
    def get_canvas_pos(self, size, val, min, max):
        """
        Compute the screen location at which a sample value would be drawn.
        ``size`` is the width or height of the chart, in points.
        ``val`` is the sample value.
        ``min`` and ``max`` are the minimum and maximum sample values that
        are to be displayed over the length of ``size``.

        For example, suppose the width of a chart is 200 points and the
        minimum and maximum X values in the sample data are 100 and 150
        respectively. When Pychart wants to draw a sample point at the X
        value of 120, it will call
        area.T.x_coord.get_canvas_pos(size = 200, val = 120, min = 100, max = 150).
        """
        raise Exception

    def get_tics(self, min, max, interval):
        """Generate the list of places for drawing tick marks."""
        raise Exception

    def get_min_max(self, min, max, interval):
        """Compute the min/max values to be displayed in the chart.
        Parameters ``min`` and ``max`` are the minimum and maximum values
        of the sample data passed to the plots. Parameter ``interval`` is
        the value of attribute area.T.x_grid_interval (or y_grid_interval).
        It is None if these attributes are non-specified.

        This method must return tuple (dmin, dmax, dinterval).
        dmin should be ``min`` rounded down to some good number.
        dmax should be ``max`` rounded up to some good number.
        dinterval should be ``interval`` if it is non-None. Otherwise, the
        method must compute some good value.
        """

        raise Exception



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import types


class IntervalTypeClass:
    def typeCheck(self, val):
        if type(val) in (int, int, float,
                         types.FunctionType):
            return None
        return "Expecting a number or a function, but received '%s'", val

    def typeDescription(self):
        return "A number or function"

    def varDescription(self, name):
        return "A number or function"


IntervalType = IntervalTypeClass()



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
__all__ = ["axis", "area", "basecanvas", "canvas",
           "line_plot", "pie_plot", "rose_plot", "tick_mark",
           "bar_plot", "chart_data", "arrow", "text_box", "color", "font",
           "fill_style", "error_bar", "range_plot", "chart_object",
           "line_style", "legend", "pychart_util", "theme", "scaling",
           "zap", "coord", "linear_coord", "log_coord",
           "category_coord", "afm", "interval_bar_plot"]



#!/usr/bin/env python

import os
from distutils.core import setup
from distutils.command.install import INSTALL_SCHEMES

data_files = []
root_dir = os.path.dirname(__file__)
if root_dir != '':
    os.chdir(root_dir)

# collect a list of all non python files
for dirpath, dirnames, filenames in os.walk('Purr'):
    dirnames[:] = [d for d in dirnames if not d.startswith('.') and d != '__pycache__']
    if filenames and '__init__.py' not in filenames:
        data_files.append([dirpath, [os.path.join(dirpath, f) for f in filenames]])

# Tell distutils not to put the data_files in platform-specific installation
# locations. See here for an explanation:
# http://groups.google.com/group/comp.lang.python/browse_thread/thread/35ec7b2fed36eaec/2105ee4d9e8042cb
for scheme in list(INSTALL_SCHEMES.values()):
        scheme['data'] = scheme['purelib']

setup(name='purr',
      version='1.5.1',
      description='Data reduction logging tool, Useful for remembering reductions',
      author='Oleg Smirnov',
      author_email='Oleg Smirnov <osmirnov@gmail.com>',
      url='https://github.com/ska-sa/purr',
      packages=['Purr', 'Purr/Plugins', 'Purr/Plugins/local_pychart', 'Purr/Plugins/local_pychart/afm'],
      install_requires=['kittens', 'pillow', 'scipy', 'astropy', 'future'],  # 'PyQt4',
      scripts=['Purr/purr.py', 'Purr/purr'],
      data_files=data_files,
     )
